#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple
from logging import getLogger

from py2neo.vendor.interchange.packstream import Structure

from py2neo.client import Hydrant
from py2neo.compat import Sequence, Mapping, integer_types, string_types
from py2neo.matching import RelationshipMatcher


log = getLogger(__name__)


INT64_MIN = -(2 ** 63)
INT64_MAX = 2 ** 63 - 1


class JSONHydrant(Hydrant):

    unbound_relationship = namedtuple("UnboundRelationship", ["id", "type", "properties"])

    def __init__(self, graph):
        self.graph = graph
        self.hydration_functions = {}

    @classmethod
    def _uri_to_id(cls, uri):
        """ Utility function to convert entity URIs into numeric identifiers.
        """
        _, _, identity = uri.rpartition("/")
        return int(identity)

    @classmethod
    def json_to_packstream(cls, data):
        """ This converts from JSON format into PackStream prior to
        proper hydration. This code needs to die horribly in a freak
        yachting accident.
        """
        # TODO: other partial hydration
        if "self" in data:
            if "type" in data:
                return Structure(ord(b"R"),
                                 cls._uri_to_id(data["self"]),
                                 cls._uri_to_id(data["start"]),
                                 cls._uri_to_id(data["end"]),
                                 data["type"],
                                 data["data"])
            else:
                return Structure(ord(b"N"),
                                 cls._uri_to_id(data["self"]),
                                 data["metadata"]["labels"],
                                 data["data"])
        elif "nodes" in data and "relationships" in data:
            nodes = [Structure(ord(b"N"), i, None, None) for i in map(cls._uri_to_id, data["nodes"])]
            relps = [Structure(ord(b"r"), i, None, None) for i in map(cls._uri_to_id, data["relationships"])]
            seq = [i // 2 + 1 for i in range(2 * len(data["relationships"]))]
            for i, direction in enumerate(data["directions"]):
                if direction == "<-":
                    seq[2 * i] *= -1
            return Structure(ord(b"P"), nodes, relps, seq)
        else:
            # from warnings import warn
            # warn("Map literals returned over the Neo4j HTTP interface are ambiguous "
            #      "and may be unintentionally hydrated as graph objects")
            return data

    def hydrate_list(self, values):
        """ Convert JSON values into native values. This is the other half
        of the HTTP hydration process, and is basically a copy of the
        Bolt/PackStream hydration code. It needs to be combined with the
        code in `json_to_packstream` so that hydration is done in a single
        pass.
        """
        assert isinstance(values, list)
        for i, value in enumerate(values):
            if isinstance(value, (list, dict, Structure)):
                values[i] = self.hydrate_object(value)
        return values

    def hydrate_object(self, obj):
        log.debug("Hydrating object %r", obj)
        from py2neo.data import Node, Relationship, Path
        if isinstance(obj, Structure):
            tag = obj.tag
            fields = obj.fields
            if tag == ord(b"N"):
                obj = Node.ref(self.graph, fields[0])
                if fields[1] is not None:
                    obj._remote_labels = frozenset(fields[1])
                    obj.clear_labels()
                    obj.update_labels(fields[1])
                if fields[2] is not None:
                    obj.clear()
                    obj.update(self.hydrate_object(fields[2]))
                return obj
            elif tag == ord(b"R"):
                start_node = Node.ref(self.graph, fields[1])
                end_node = Node.ref(self.graph, fields[2])
                obj = Relationship.ref(self.graph, fields[0], start_node, fields[3], end_node)
                if fields[4] is not None:
                    obj.clear()
                    obj.update(self.hydrate_object(fields[4]))
                return obj
            elif tag == ord(b"P"):
                # Herein lies a dirty hack to retrieve missing relationship
                # detail for paths received over HTTP.
                nodes = [self.hydrate_object(node) for node in fields[0]]
                u_rels = []
                typeless_u_rel_ids = []
                for r in fields[1]:
                    u_rel = self.unbound_relationship(*map(self.hydrate_object, r))
                    assert u_rel.type is None
                    typeless_u_rel_ids.append(u_rel.id)
                    u_rels.append(u_rel)
                if typeless_u_rel_ids:
                    r_dict = {r.identity: r for r in RelationshipMatcher(self.graph).get(typeless_u_rel_ids)}
                    for i, u_rel in enumerate(u_rels):
                        if u_rel.type is None:
                            u_rels[i] = self.unbound_relationship(
                                u_rel.id,
                                type(r_dict[u_rel.id]).__name__,
                                u_rel.properties
                            )
                sequence = fields[2]
                return Path.hydrate(self.graph, nodes, u_rels, sequence)
            else:
                try:
                    f = self.hydration_functions[tag]
                except KeyError:
                    # If we don't recognise the structure type, just return it as-is
                    return obj
                else:
                    return f(*map(self.hydrate_object, obj.fields))
        elif isinstance(obj, list):
            return list(map(self.hydrate_object, obj))
        elif isinstance(obj, dict):
            return {key: self.hydrate_object(value) for key, value in obj.items()}
        else:
            return obj

    def dehydrate(self, data, version=None):
        """ Dehydrate to JSON.
        """
        if data is None or data is True or data is False or isinstance(data, float) or isinstance(data, string_types):
            return data
        elif isinstance(data, integer_types):
            if data < INT64_MIN or data > INT64_MAX:
                raise ValueError("Integers must be within the signed 64-bit range")
            return data
        elif isinstance(data, bytearray):
            return list(data)
        elif isinstance(data, Mapping):
            d = {}
            for key in data:
                if not isinstance(key, string_types):
                    raise TypeError("Dictionary keys must be strings")
                d[key] = self.dehydrate(data[key])
            return d
        elif isinstance(data, Sequence):
            return list(map(self.dehydrate, data))
        else:
            raise TypeError("Neo4j does not support JSON parameters of type %s" % type(data).__name__)


def dehydrate(data):
    """ Dehydrate to JSON.
    """
    if data is None or data is True or data is False or isinstance(data, float) or isinstance(data, string_types):
        return data
    elif isinstance(data, integer_types):
        if data < INT64_MIN or data > INT64_MAX:
            raise ValueError("Integers must be within the signed 64-bit range")
        return data
    elif isinstance(data, bytearray):
        return list(data)
    elif isinstance(data, Mapping):
        d = {}
        for key in data:
            if not isinstance(key, string_types):
                raise TypeError("Dictionary keys must be strings")
            d[key] = dehydrate(data[key])
        return d
    elif isinstance(data, Sequence):
        return list(map(dehydrate, data))
    else:
        raise TypeError("Neo4j does not support JSON parameters of type %s" % type(data).__name__)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import, print_function, unicode_literals


"""
Provides integration with `pandas <https://pandas.pydata.org/>`_.

.. note::
   This module requires pandas to be installed, and will raise a
   warning if this is not available.

Example:

    >>> from py2neo import Graph
    >>> from py2neo.integration.pandas import cursor_to_data_frame
    >>> graph = Graph()
    >>> df = cursor_to_data_frame(graph.query("MATCH (a:Person) RETURN a.name AS name, a.born AS born"))
    >>> df
                       name    born
    0          Keanu Reeves  1964.0
    1      Carrie-Anne Moss  1967.0
    2    Laurence Fishburne  1961.0
    3          Hugo Weaving  1960.0
    4       Lilly Wachowski  1967.0
    ..                  ...     ...
    128      Penny Marshall  1943.0
    129         Paul Blythe     NaN
    130        Angela Scope     NaN
    131    Jessica Thompson     NaN
    132      James Thompson     NaN

    [133 rows x 2 columns]

"""


from warnings import warn

try:
    # noinspection PyPackageRequirements
    from pandas import DataFrame, Series
except ImportError:
    warn("The py2neo.integration.pandas module expects pandas to be "
         "installed but it does not appear to be available.")
    raise

from py2neo.bulk import create_nodes, merge_nodes


def cursor_to_series(cursor, field=0, index=None, dtype=None):
    """ Consume and extract one field of the entire result as a
    `pandas.Series <https://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`_.

    :param cursor:
    :param field:
    :param index:
    :param dtype:
    :returns: `Series
        <https://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__ object.
    """
    return Series([record[field] for record in cursor], index=index, dtype=dtype)


def cursor_to_data_frame(cursor, index=None, columns=None, dtype=None):
    """ Consume and extract the entire result as a
    `pandas.DataFrame <https://pandas.pydata.org/pandas-docs/stable/dsintro.html#dataframe>`_.

    :param cursor:
    :param index: Index to use for resulting frame.
    :param columns: Column labels to use for resulting frame.
    :param dtype: Data type to force.
    :returns: `DataFrame
        <https://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__ object.
    """
    return DataFrame(list(map(dict, cursor)), index=index, columns=columns, dtype=dtype)


def create_nodes_from_data_frame(tx, df, labels=None):
    """ Create nodes from a DataFrame.

    This function wraps the :func:`py2neo.bulk.create_nodes` function,
    allowing a `DataFrame <https://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__
    object to be passed in place of the regular `data` argument.

    :param tx:
    :param df:
    :param labels:
    :return:
    """
    create_nodes(tx, df.itertuples(index=False, name=None),
                 labels=labels, keys=df.keys())


def merge_nodes_from_data_frame(tx, df, merge_key, labels=None, preserve=None):
    """ Merge nodes from a DataFrame.

    This function wraps the :func:`py2neo.bulk.merge_nodes` function,
    allowing a `DataFrame <https://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__
    object to be passed in place of the regular `data` argument.

    :param tx:
    :param df:
    :param merge_key:
    :param labels:
    :param preserve:
    :return:
    """
    merge_nodes(tx, df.itertuples(index=False, name=None), merge_key,
                labels=labels, keys=df.keys(), preserve=preserve)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from py2neo.cypher import cypher_escape


class ProcedureLibrary(object):
    """ Accessor for listing and calling procedures.

    This object is typically constructed and accessed via the
    :meth:`.Graph.call` attribute. See the documentation for that
    attribute for usage information.

    *New in version 2020.0.*
    """

    def __init__(self, graph):
        self.graph = graph

    def __getattr__(self, name):
        return Procedure(self.graph, name)

    def __getitem__(self, name):
        return Procedure(self.graph, name)

    def __dir__(self):
        return list(self)

    def __iter__(self):
        proc = Procedure(self.graph, "dbms.procedures")
        for record in proc(keys=["name"]):
            yield record[0]

    def __call__(self, procedure, *args):
        """ Call a procedure by name.

        For example:
            >>> from py2neo import Graph
            >>> g = Graph()
            >>> g.call("dbms.components")
             name         | versions  | edition
            --------------|-----------|-----------
             Neo4j Kernel | ['4.0.2'] | community

        :param procedure: fully qualified procedure name
        :param args: positional arguments to pass to the procedure
        :returns: :class:`.Cursor` object wrapping the result
        """
        return Procedure(self.graph, procedure)(*args)


class Procedure(object):
    """ Represents an individual procedure.

    *New in version 2020.0.*
    """

    def __init__(self, graph, name):
        self.graph = graph
        self.name = name

    def __getattr__(self, name):
        return Procedure(self.graph, self.name + "." + name)

    def __getitem__(self, name):
        return Procedure(self.graph, self.name + "." + name)

    def __dir__(self):
        proc = Procedure(self.graph, "dbms.procedures")
        prefix = self.name + "."
        return [record[0][len(prefix):] for record in proc(keys=["name"])
                if record[0].startswith(prefix)]

    def __call__(self, *args, **kwargs):
        """ Call a procedure by name.

        For example:
            >>> from py2neo import Graph
            >>> g = Graph()
            >>> g.call("dbms.components")
             name         | versions  | edition
            --------------|-----------|-----------
             Neo4j Kernel | ['4.0.2'] | community

        :param procedure: fully qualified procedure name
        :param args: positional arguments to pass to the procedure
        :returns: :class:`.Cursor` object wrapping the result
        """
        procedure_name = ".".join(cypher_escape(part) for part in self.name.split("."))
        arg_list = [(str(i), arg) for i, arg in enumerate(args)]
        cypher = "CALL %s(%s)" % (procedure_name, ", ".join("$" + a[0] for a in arg_list))
        keys = kwargs.get("keys")
        if keys:
            cypher += " YIELD %s" % ", ".join(keys)
        return self.graph.run(cypher, dict(arg_list))



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "Neo4jError",
    "ClientError",
    "DatabaseError",
    "TransientError",
    "ConnectionUnavailable",
    "ConnectionBroken",
    "ConnectionLimit",
    "ServiceUnavailable",
    "WriteServiceUnavailable",
    "ProtocolError",
]


class Neo4jError(Exception):
    """ Base exception class for modelling error status codes returned
    by Neo4j.

    For details of the status codes available, visit:
    https://neo4j.com/docs/status-codes/current/

    """

    @classmethod
    def hydrate(cls, data):
        code = data.get("code")
        message = data.get("message")
        return cls(message, code)

    @classmethod
    def split_code(cls, code):
        """ Splits a status code, returning a 3-tuple of
        classification, category and title.
        """
        try:
            parts = code.split(".")
        except AttributeError:
            raise ValueError(code)
        else:
            if len(parts) == 4 and parts[0] == "Neo":
                return parts[1], parts[2], parts[3]
            else:
                raise ValueError(code)

    def __new__(cls, message, code):
        classification, _, _ = cls.split_code(code)
        if classification == "ClientError":
            return Exception.__new__(ClientError)
        elif classification == "DatabaseError":
            return Exception.__new__(DatabaseError)
        elif classification == "TransientError":
            return Exception.__new__(TransientError)
        else:
            return Exception.__new__(cls)

    def __init__(self, message, code):
        super(Neo4jError, self).__init__(message)
        self.__code = code
        self.__classification, self.__category, self.__title = self.split_code(self.__code)

    def __str__(self):
        return "[%s.%s] %s" % (self.category, self.title, super(Neo4jError, self).__str__())

    @property
    def code(self):
        return self.__code

    @property
    def classification(self):
        return self.__classification

    @property
    def category(self):
        return self.__category

    @property
    def title(self):
        return self.__title

    @property
    def message(self):
        return self.args[0]

    def should_retry(self):
        return False

    def should_invalidate_routing_table(self):
        return False


class ClientError(Neo4jError):

    def should_invalidate_routing_table(self):
        return self.category == "Cluster" and self.title == "NotALeader"

    def should_retry(self):
        return self.category == "Cluster" and self.title == "NotALeader"


class DatabaseError(Neo4jError):
    pass


class TransientError(Neo4jError):

    def should_retry(self):
        return True


class ConnectionUnavailable(Exception):
    """ Raised when a connection cannot be acquired.
    """


class ConnectionBroken(Exception):
    """ Raised when a connection breaks during use.
    """


class ConnectionLimit(Exception):
    """ Raised when no further connections are available
    due to a configured resource limit.
    """


class ServiceUnavailable(Exception):
    """ Raised when no read or write service is available from a
    Neo4j installation.
    """

    def __init__(self, *args):
        super(ServiceUnavailable, self).__init__(*args)


class WriteServiceUnavailable(ServiceUnavailable):
    """ Raised when no write service is available from a
    Neo4j installation, but a read service may still be available.
    """

    def __init__(self, *args):
        super(WriteServiceUnavailable, self).__init__(*args)


class ProtocolError(Exception):
    """ Raised when a protocol violation or other unrecoverable
    protocol error occurs. These errors cannot be remedied
    automatically, and may result from a bug in the driver or server
    software.
    """
    # TODO: add hints for users when they see this error



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module provides a convenience socket wrapper class
(:class:`.Wire`) for low-level network communication.
"""


from socket import SHUT_WR

from monotonic import monotonic
from six import raise_from

from py2neo.addressing import Address
from py2neo.compat import BaseRequestHandler


class Wire(object):
    """ Buffered socket wrapper for reading and writing bytes.
    """

    __closed = False

    __broken = False

    @classmethod
    def open(cls, address, timeout=None, keep_alive=False, on_broken=None):
        """ Open a connection to a given network :class:`.Address`.

        :param address:
        :param timeout:
        :param keep_alive:
        :param on_broken: callback for when the wire is broken after a
            successful connection has first been established (this does
            not trigger if the connection never opens successfully)
        :returns: :class:`.Wire` object
        :raises WireError: if connection fails to open
        """
        from socket import socket, SOL_SOCKET, SO_KEEPALIVE
        address = Address(address)
        s = socket(family=address.family)
        if keep_alive:
            s.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
        s.settimeout(timeout)
        try:
            s.connect(address)
        except (IOError, OSError) as error:
            raise_from(WireError("Cannot connect to %r" % (address,)), error)
        return cls(s, on_broken=on_broken)

    def __init__(self, s, on_broken=None):
        s.settimeout(None)  # ensure wrapped socket is in blocking mode
        self.__socket = s
        self.__active_time = monotonic()
        self.__bytes_received = 0
        self.__bytes_sent = 0
        self.__input = bytearray()
        self.__input_len = 0
        self.__output = bytearray()
        self.__on_broken = on_broken

    def secure(self, verify=True, hostname=None):
        """ Apply a layer of security onto this connection.
        """
        from ssl import SSLContext, SSLError
        try:
            # noinspection PyUnresolvedReferences
            from ssl import PROTOCOL_TLS
        except ImportError:
            from ssl import PROTOCOL_SSLv23
            context = SSLContext(PROTOCOL_SSLv23)
        else:
            context = SSLContext(PROTOCOL_TLS)
        if verify:
            from ssl import CERT_REQUIRED
            context.verify_mode = CERT_REQUIRED
            context.check_hostname = bool(hostname)
        else:
            from ssl import CERT_NONE
            context.verify_mode = CERT_NONE
        context.load_default_certs()
        try:
            self.__socket = context.wrap_socket(self.__socket, server_hostname=hostname)
        except (IOError, OSError) as error:
            # TODO: add connection failure/diagnostic callback
            if error.errno == 0:
                raise BrokenWireError("Peer closed connection during TLS handshake; "
                                      "server may not be configured for secure connections")
            else:
                raise WireError("Unable to establish secure connection with remote peer")
        else:
            self.__active_time = monotonic()

    def read(self, n):
        """ Read bytes from the network.
        """
        while self.__input_len < n:
            required = n - self.__input_len
            requested = max(required, 16384)
            try:
                received = self.__socket.recv(requested)
            except (IOError, OSError):
                self.__mark_broken("Wire broken")
            else:
                if received:
                    self.__active_time = monotonic()
                    new_bytes_received = len(received)
                    self.__input.extend(received)
                    self.__input_len += new_bytes_received
                    self.__bytes_received += new_bytes_received
                else:
                    self.__mark_broken("Network read incomplete "
                                       "(received %d of %d bytes)" %
                                       (self.__input_len, n))
        data = self.__input[:n]
        self.__input[:n] = []
        self.__input_len -= n
        return data

    def peek(self):
        """ Return any buffered unread data.
        """
        return self.__input

    def write(self, b):
        """ Write bytes to the output buffer.
        """
        self.__output.extend(b)

    def send(self, final=False):
        """ Send the contents of the output buffer to the network.
        """
        if self.__closed:
            raise WireError("Closed")
        sent = 0
        while self.__output:
            try:
                n = self.__socket.send(self.__output)
            except (IOError, OSError):
                self.__mark_broken("Wire broken")
            else:
                self.__active_time = monotonic()
                self.__bytes_sent += n
                self.__output[:n] = []
                sent += n
        if final:
            try:
                self.__socket.shutdown(SHUT_WR)
            except (IOError, OSError):
                self.__mark_broken("Wire broken")
        return sent

    def close(self):
        """ Close the connection.
        """
        try:
            self.__socket.close()
        except (IOError, OSError):
            self.__mark_broken("Wire broken")
        else:
            self.__closed = True

    @property
    def closed(self):
        """ Flag indicating whether this connection has been closed locally.
        """
        return self.__closed

    @property
    def broken(self):
        """ Flag indicating whether this connection has been closed remotely.
        """
        return self.__broken

    @property
    def local_address(self):
        """ The local :class:`.Address` to which this connection is bound.
        """
        return Address(self.__socket.getsockname())

    @property
    def remote_address(self):
        """ The remote :class:`.Address` to which this connection is bound.
        """
        return Address(self.__socket.getpeername())

    @property
    def bytes_sent(self):
        return self.__bytes_sent

    @property
    def bytes_received(self):
        return self.__bytes_received

    def __mark_broken(self, message):
        idle_time = monotonic() - self.__active_time
        message += (" after %.01fs idle (%r bytes sent, "
                    "%r bytes received)" % (idle_time,
                                            self.__bytes_sent,
                                            self.__bytes_received))
        if callable(self.__on_broken):
            self.__on_broken(message)
        self.__broken = True
        raise BrokenWireError(message, idle_time=idle_time,
                              bytes_sent=self.__bytes_sent,
                              bytes_received=self.__bytes_received)


class WireRequestHandler(BaseRequestHandler):
    """ Base handler for use with the `socketserver` module that wraps
    the request attribute as a :class:`.Wire` object.
    """

    __wire = None

    @property
    def wire(self):
        if self.__wire is None:
            self.__wire = Wire(self.request)
        return self.__wire


class WireError(OSError):
    """ Raised when a connection error occurs.

    :param idle_time:
    :param bytes_sent:
    :param bytes_received:
    """

    def __init__(self, *args, **kwargs):
        super(WireError, self).__init__(*args)
        self.idle_time = kwargs.get("idle_time", None)
        self.bytes_sent = kwargs.get("bytes_sent", 0)
        self.bytes_received = kwargs.get("bytes_received", 0)


class BrokenWireError(WireError):
    """ Raised when a connection is broken by the network or remote peer.
    """



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "Cursor",
    "Record",
    "CypherExpression",
    "cypher_escape",
    "cypher_join",
    "cypher_repr",
    "cypher_str",
]

from functools import reduce
from io import StringIO
from operator import xor as xor_operator

from py2neo.cypher.encoding import CypherEncoder
from py2neo.compat import Mapping, string_types, unicode_types, ustr


class Cursor(object):
    """ A `Cursor` is a navigator for a stream of records.

    A cursor can be thought of as a window onto an underlying data
    stream. All cursors in py2neo are "forward-only", meaning that
    navigation starts before the first record and may proceed only in a
    forward direction.

    It is not generally necessary for application code to instantiate a
    cursor directly as one will be returned by any Cypher execution method.
    However, cursor creation requires only a :class:`.DataSource` object
    which contains the logic for how to access the source data that the
    cursor navigates.

    Many simple cursor use cases require only the :meth:`.forward` method
    and the :attr:`.current` attribute. To navigate through all available
    records, a `while` loop can be used::

        while cursor.forward():
            print(cursor.current["name"])

    If only the first record is of interest, a similar `if` structure will
    do the job::

        if cursor.forward():
            print(cursor.current["name"])

    To combine `forward` and `current` into a single step, use the built-in
    py:func:`next` function::

        print(next(cursor)["name"])

    Cursors are also iterable, so can be used in a loop::

        for record in cursor:
            print(record["name"])

    For queries that are expected to return only a single value within a
    single record, use the :meth:`.evaluate` method. This will return the
    first value from the next record or :py:const:`None` if neither the
    field nor the record are present::

        print(cursor.evaluate())

    """

    def __init__(self, result, hydrant=None, sample_size=3):
        self._result = result
        self._fields = self._result.fields()
        self._hydrant = hydrant
        self._current = None
        self.sample_size = sample_size

    def __repr__(self):
        preview = self.preview()
        if preview:
            return repr(preview)
        else:
            return "(No data)"

    def _repr_html_(self):
        preview = self.preview()
        if preview:
            s = StringIO()
            preview.write_html(file=s, header=True)
            return s.getvalue()
        else:
            return "(No data)"

    def __next__(self):
        if self.forward():
            return self._current
        else:
            raise StopIteration()

    # Exists only for Python 2 iteration compatibility
    next = __next__

    def __iter__(self):
        while self.forward():
            yield self._current

    def __getitem__(self, key):
        return self._current[key]

    @property
    def profile(self):
        return self._result.profile

    @property
    def current(self):
        """ Returns the current record or :py:const:`None` if no record
        has yet been selected.
        """
        return self._current

    @property
    def closed(self):
        return self._result.offline

    def keys(self):
        """ Return the field names for the records in the stream.
        """
        return self._fields

    def summary(self):
        """ Return the result summary.
        """
        return self._result.summary()

    def plan(self):
        """ Return the execution plan returned by this query, if any.
        """
        metadata = self._result.summary()
        try:
            return metadata["plan"]
        except KeyError:
            try:
                return metadata["profile"]
            except KeyError:
                return None

    def stats(self):
        """ Return the execution statistics for this query.

        This contains details of the activity undertaken by the database
        kernel for the query, such as the number of entities created or
        deleted.

        >>> from py2neo import Graph
        >>> g = Graph()
        >>> g.run("CREATE (a:Person) SET a.name = 'Alice'").stats()
        {'labels_added': 1, 'nodes_created': 1, 'properties_set': 1}

        """
        metadata = self._result.summary()
        stats = {}
        for key, value in metadata.get("stats", {}).items():
            key = key.replace("-", "_")
            if key.startswith("relationship_"):
                # hack for server bug
                key = "relationships_" + key[13:]
            stats[key] = value
        return stats

    def forward(self, amount=1):
        """ Attempt to move the cursor one position forward (or by
        another amount if explicitly specified). The cursor will move
        position by up to, but never more than, the amount specified.
        If not enough scope for movement remains, only that remainder
        will be consumed. The total amount moved is returned.

        :param amount: the amount to move the cursor
        :returns: the amount that the cursor was able to move
        """
        if amount == 0:
            return 0
        if amount < 0:
            raise ValueError("Cursor can only move forwards")
        amount = int(amount)
        moved = 0
        while moved != amount:
            values = self._result.take()
            if values is None:
                break
            if self._hydrant:
                values = self._hydrant.hydrate_list(values)
            self._current = Record(self._fields, values)
            moved += 1
        return moved

    def preview(self, limit=None):
        """ Construct a :class:`.Table` containing a preview of
        upcoming records, including no more than the given `limit`.

        :param limit: maximum number of records to include in the
            preview
        :returns: :class:`.Table` containing the previewed records
        """
        from py2neo.integration import Table
        if not limit:
            limit = self.sample_size
        elif limit < 0:
            raise ValueError("Illegal preview size")
        else:
            limit = int(limit)
        records = []
        if self._fields:
            for values in self._result.peek(limit):
                if self._hydrant:
                    values = self._hydrant.hydrate_list(values)
                records.append(values)
            return Table(records, self._fields)
        else:
            return None

    def evaluate(self, field=0):
        """ Return the value of the first field from the next record
        (or the value of another field if explicitly specified).

        This method attempts to move the cursor one step forward and,
        if successful, selects and returns an individual value from
        the new current record. By default, this value will be taken
        from the first value in that record but this can be overridden
        with the `field` argument, which can represent either a
        positional index or a textual key.

        If the cursor cannot be moved forward or if the record contains
        no values, :py:const:`None` will be returned instead.

        This method is particularly useful when it is known that a
        Cypher query returns only a single value.

        :param field: field to select value from (optional)
        :returns: value of the field or :py:const:`None`

        Example:
            >>> from py2neo import Graph
            >>> g = Graph()
            >>> g.run("MATCH (a) WHERE a.email=$x RETURN a.name", x="bob@acme.com").evaluate()
            'Bob Robertson'
        """
        if self.forward():
            try:
                return self[field]
            except IndexError:
                return None
        else:
            return None

    def data(self, *keys):
        """ Consume and extract the entire result as a list of
        dictionaries.

        ::

            >>> from py2neo import Graph
            >>> graph = Graph()
            >>> graph.run("MATCH (a:Person) RETURN a.name, a.born LIMIT 4").data()
            [{'a.born': 1964, 'a.name': 'Keanu Reeves'},
             {'a.born': 1967, 'a.name': 'Carrie-Anne Moss'},
             {'a.born': 1961, 'a.name': 'Laurence Fishburne'},
             {'a.born': 1960, 'a.name': 'Hugo Weaving'}]

        :param keys: indexes or keys of the items to include; if none
                     are provided, all values will be included
        :returns: list of dictionary of values, keyed by field name
        :raises IndexError: if an out-of-bounds index is specified
        """
        return [record.data(*keys) for record in self]

    def to_table(self):
        """ Consume and extract the entire result as a :class:`.Table`
        object.

        :return: the full query result
        """
        from py2neo.integration import Table
        return Table(self)

    def to_subgraph(self):
        """ Consume and extract the entire result as a :class:`.Subgraph`
        containing the union of all the graph structures within.

        :return: :class:`.Subgraph` object
        """
        s = None
        for record in self:
            s_ = record.to_subgraph()
            if s_ is not None:
                if s is None:
                    s = s_
                else:
                    s = s | s_
        return s

    def to_ndarray(self, dtype=None, order='K'):
        """ Consume and extract the entire result as a
        `numpy.ndarray <https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html>`_.

        .. note::
           This method requires `numpy` to be installed.

        :param dtype:
        :param order:
        :warns: If `numpy` is not installed
        :returns: `ndarray
            <https://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.html>`__ object.
        """
        from py2neo.integration.numpy import cursor_to_ndarray
        return cursor_to_ndarray(self, dtype, order)

    def to_series(self, field=0, index=None, dtype=None):
        """ Consume and extract one field of the entire result as a
        `pandas.Series <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`_.

        .. note::
           This method requires `pandas` to be installed.

        :param field:
        :param index:
        :param dtype:
        :warns: If `pandas` is not installed
        :returns: `Series
            <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__ object.
        """
        from py2neo.integration.pandas import cursor_to_series
        return cursor_to_series(self, field, index, dtype)

    def to_data_frame(self, index=None, columns=None, dtype=None):
        """ Consume and extract the entire result as a
        `pandas.DataFrame <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#dataframe>`_.

        ::

            >>> from py2neo import Graph
            >>> graph = Graph()
            >>> graph.run("MATCH (a:Person) RETURN a.name, a.born LIMIT 4").to_data_frame()
               a.born              a.name
            0    1964        Keanu Reeves
            1    1967    Carrie-Anne Moss
            2    1961  Laurence Fishburne
            3    1960        Hugo Weaving

        .. note::
           This method requires `pandas` to be installed.

        :param index: Index to use for resulting frame.
        :param columns: Column labels to use for resulting frame.
        :param dtype: Data type to force.
        :warns: If `pandas` is not installed
        :returns: `DataFrame
            <http://pandas.pydata.org/pandas-docs/stable/dsintro.html#series>`__ object.
        """
        from py2neo.integration.pandas import cursor_to_data_frame
        return cursor_to_data_frame(self, index, columns, dtype)

    def to_matrix(self, mutable=False):
        """ Consume and extract the entire result as a
        `sympy.Matrix <http://docs.sympy.org/latest/tutorial/matrices.html>`_.

        .. note::
           This method requires `sympy` to be installed.

        :param mutable:
        :returns: `Matrix
            <http://docs.sympy.org/latest/tutorial/matrices.html>`_ object.
        """
        from py2neo.integration.sympy import cursor_to_matrix
        return cursor_to_matrix(self, mutable)


class Record(tuple, Mapping):
    """ A :class:`.Record` object holds an ordered, keyed collection of
    values. It is in many ways similar to a :class:`namedtuple` but
    allows field access only through bracketed syntax, and provides
    more functionality. :class:`.Record` extends both :class:`tuple`
    and :class:`Mapping`.

    .. describe:: record[index]
                  record[key]

        Return the value of *record* with the specified *key* or *index*.

    .. describe:: len(record)

        Return the number of fields in *record*.

    .. describe:: dict(record)

        Return a `dict` representation of *record*.

    """

    __keys = None

    def __new__(cls, keys, values):
        inst = tuple.__new__(cls, values)
        inst.__keys = keys
        return inst

    def __repr__(self):
        return "Record({%s})" % ", ".join("%r: %r" % (field, self[i])
                                          for i, field in enumerate(self.__keys))

    def __str__(self):
        return "\t".join(map(repr, (self[i] for i, _ in enumerate(self.__keys))))

    def __eq__(self, other):
        return dict(self) == dict(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return reduce(xor_operator, map(hash, self.items()))

    def __getitem__(self, key):
        if isinstance(key, slice):
            keys = self.__keys[key]
            values = super(Record, self).__getitem__(key)
            return self.__class__(zip(keys, values))
        index = self.index(key)
        if 0 <= index < len(self):
            return super(Record, self).__getitem__(index)
        else:
            return None

    def __getslice__(self, start, stop):
        key = slice(start, stop)
        keys = self.__keys[key]
        values = tuple(self)[key]
        return self.__class__(zip(keys, values))

    def get(self, key, default=None):
        """ Obtain a single value from the record by index or key. If the
        specified item does not exist, the default value is returned.

        :param key: index or key
        :param default: default value to be returned if `key` does not exist
        :return: selected value
        """
        try:
            index = self.__keys.index(ustr(key))
        except ValueError:
            return default
        if 0 <= index < len(self):
            return super(Record, self).__getitem__(index)
        else:
            return default

    def index(self, key):
        """ Return the index of the given item.
        """
        from six import integer_types, string_types
        if isinstance(key, integer_types):
            if 0 <= key < len(self.__keys):
                return key
            raise IndexError(key)
        elif isinstance(key, string_types):
            try:
                return self.__keys.index(key)
            except ValueError:
                raise KeyError(key)
        else:
            raise TypeError(key)

    def keys(self):
        """ Return the keys of the record.

        :return: list of key names
        """
        return list(self.__keys)

    def values(self, *keys):
        """ Return the values of the record, optionally filtering to
        include only certain values by index or key.

        :param keys: indexes or keys of the items to include; if none
                     are provided, all values will be included
        :return: list of values
        """
        if keys:
            d = []
            for key in keys:
                try:
                    i = self.index(key)
                except KeyError:
                    d.append(None)
                else:
                    d.append(self[i])
            return d
        return list(self)

    def items(self, *keys):
        """ Return the fields of the record as a list of key and value tuples

        :param keys: indexes or keys of the items to include; if none
                     are provided, all values will be included
        :return: list of (key, value) tuples
        """
        if keys:
            d = []
            for key in keys:
                try:
                    i = self.index(key)
                except KeyError:
                    d.append((key, None))
                else:
                    d.append((self.__keys[i], self[i]))
            return d
        return list((self.__keys[i], super(Record, self).__getitem__(i)) for i in range(len(self)))

    def data(self, *keys):
        """ Return the keys and values of this record as a dictionary,
        optionally including only certain values by index or key. Keys
        provided that do not exist within the record will be included
        but with a value of :py:const:`None`; indexes provided
        that are out of bounds will trigger an :exc:`IndexError`.

        :param keys: indexes or keys of the items to include; if none
                     are provided, all values will be included
        :return: dictionary of values, keyed by field name
        :raises: :exc:`IndexError` if an out-of-bounds index is specified
        """
        if keys:
            d = {}
            for key in keys:
                try:
                    i = self.index(key)
                except KeyError:
                    d[key] = None
                else:
                    d[self.__keys[i]] = self[i]
            return d
        return dict(self)

    def to_subgraph(self):
        """ Return a :class:`.Subgraph` containing the union of all the
        graph structures within this :class:`.Record`.

        :return: :class:`.Subgraph` object
        """
        from py2neo.data import Subgraph
        s = None
        for value in self.values():
            if isinstance(value, Subgraph):
                if s is None:
                    s = value
                else:
                    s = s | value
        return s


class CypherExpression(object):

    def __init__(self, value):
        self.__value = value

    @property
    def value(self):
        return self.__value


def cypher_escape(identifier):
    """ Return a Cypher identifier, with escaping if required.

    Simple Cypher identifiers, which just contain alphanumerics
    and underscores, can be represented as-is in expressions.
    Any which contain more esoteric characters, such as spaces
    or punctuation, must be escaped in backticks. Backticks
    themselves are escaped by doubling.

    ::

        >>> cypher_escape("simple_identifier")
        'simple_identifier'
        >>> cypher_escape("identifier with spaces")
        '`identifier with spaces`'
        >>> cypher_escape("identifier with `backticks`")
        '`identifier with ``backticks```'

    Identifiers are used in Cypher to denote named values, labels,
    relationship types and property keys. This function will typically
    be used to construct dynamic Cypher queries in places where
    parameters cannot be used.

        >>> "MATCH (a:{label}) RETURN id(a)".format(label=cypher_escape("Employee of the Month"))
        'MATCH (a:`Employee of the Month`) RETURN id(a)'

    :param identifier: any non-empty string
    """
    if not isinstance(identifier, string_types):
        raise TypeError(type(identifier).__name__)
    encoder = CypherEncoder()
    return encoder.encode_key(identifier)


def cypher_join(*clauses, **parameters):
    """ Join multiple Cypher clauses, returning a (query, parameters)
    tuple. Each clause may either be a simple string query or a
    (query, parameters) tuple. Additional `parameters` may also be
    supplied as keyword arguments.

    :param clauses:
    :param parameters:
    :return: (query, parameters) tuple
    """
    query = []
    params = {}
    for clause in clauses:
        if clause is None:
            continue
        if isinstance(clause, tuple):
            try:
                q, p = clause
            except ValueError:
                raise ValueError("Expected query or (query, parameters) tuple "
                                 "for clause %r" % clause)
        else:
            q = clause
            p = None
        query.append(q)
        if p:
            params.update(p)
    params.update(parameters)
    return "\n".join(query), params


def cypher_repr(value, **kwargs):
    """ Return the Cypher representation of a value.

    This function attempts to convert the supplied value into a Cypher
    literal form, as used in expressions.

    """
    encoder = CypherEncoder(**kwargs)
    return encoder.encode_value(value)


def cypher_str(value, **kwargs):
    """ Convert a Cypher value to a Python Unicode string.

    This function converts the supplied value into a string form, as
    used for human-readable output. This is generally identical to
    :meth:`.cypher_repr` except for with string values, which are
    returned as-is, instead of being enclosed in quotes with certain
    characters escaped.

    """
    if isinstance(value, unicode_types):
        return value
    elif isinstance(value, string_types):
        return value.decode(kwargs.get("encoding", "utf-8"))
    else:
        return cypher_repr(value, **kwargs)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module provides classes for modelling IP addresses, based on tuples.
"""


from socket import AF_INET, AF_INET6

from py2neo.compat import xstr


class Address(tuple):
    """ Address of a machine on a network.
    """

    @classmethod
    def parse(cls, s, default_host=None, default_port=None):
        s = xstr(s)
        if s.startswith("["):
            # IPv6
            host, _, port = s[1:].rpartition("]")
            port = port.lstrip(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return cls((host or default_host or "localhost",
                        port or default_port or 0, 0, 0))
        else:
            # IPv4
            host, _, port = s.partition(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return cls((host or default_host or "localhost",
                        port or default_port or 0))

    def __new__(cls, iterable):
        if isinstance(iterable, cls):
            return iterable
        n_parts = len(iterable)
        inst = tuple.__new__(cls, iterable)
        if n_parts == 2:
            inst.__class__ = IPv4Address
        elif n_parts == 4:
            inst.__class__ = IPv6Address
        else:
            raise ValueError("Addresses must consist of either "
                             "two parts (IPv4) or four parts (IPv6)")
        return inst

    #: Address family (AF_INET or AF_INET6)
    family = None

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, tuple(self))

    @property
    def host(self):
        return self[0]

    @property
    def port(self):
        return self[1]

    @property
    def port_number(self):
        from socket import getservbyname
        if self.port == "bolt":
            # Special case, just because. The regular /etc/services
            # file doesn't contain this, but it can be found in
            # /usr/share/nmap/nmap-services if nmap is installed.
            from py2neo import DEFAULT_BOLT_PORT
            return DEFAULT_BOLT_PORT
        try:
            return getservbyname(self.port)
        except (IOError, OSError, TypeError):
            # OSError: service/proto not found
            # TypeError: getservbyname() argument 1 must be str, not X
            try:
                return int(self.port)
            except (TypeError, ValueError) as e:
                raise type(e)("Unknown port value %r" % self.port)


class IPv4Address(Address):
    """ Address subclass, specifically for IPv4 addresses.
    """

    family = AF_INET

    def __str__(self):
        return "{}:{}".format(*self)


class IPv6Address(Address):
    """ Address subclass, specifically for IPv6 addresses.
    """

    family = AF_INET6

    def __str__(self):
        return "[{}]:{}".format(*self)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = [
    "Property",
    "Label",
    "Related",
    "RelatedTo",
    "RelatedFrom",
    "RelatedObjects",
    "ModelType",
    "Model", "Model",
    "ModelMatch",
    "ModelMatcher",
    "Repository",
]

from collections import OrderedDict

from py2neo.vendor.interchange.collections import PropertyDict
from py2neo.vendor.interchange.text import Words

from py2neo.compat import metaclass, deprecated
from py2neo.cypher import cypher_escape
from py2neo.data import Node
from py2neo.database import Graph
from py2neo.matching import NodeMatch, NodeMatcher


OUTGOING = 1
UNDIRECTED = 0
INCOMING = -1


class Property(object):
    """ Property definition for a :class:`.Model`.

    Attributes:
        key: The name of the node property within the database.
        default: The default value for the property, if it would
                 otherwise be :const:`None`.
    """

    def __init__(self, key=None, default=None):
        """ Initialise a property definition.

        Args:
            key: The name of the node property within the database. If
                 omitted, the name of the class attribute is used.
            default: The default value for the property, if it would
                     otherwise be :const:`None`.
        """
        self.key = key
        self.default = default

    def __get__(self, instance, owner):
        value = instance.__node__[self.key]
        if value is None:
            value = self.default
        return value

    def __set__(self, instance, value):
        instance.__node__[self.key] = value

    def __repr__(self):
        args = OrderedDict()
        if self.key is not None:
            args["key"] = self.key
        if self.default is not None:
            args["default"] = self.default
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join("%s=%r" % arg for arg in args.items()))


class Label(object):
    """ Label definition for a :class:`.Model`.

    Labels are toggleable tags applied to an object that can be used as type
    information or other forms of classification.
    """

    def __init__(self, name=None):
        self.name = name

    def __get__(self, instance, owner):
        return instance.__node__.has_label(self.name)

    def __set__(self, instance, value):
        if value:
            instance.__node__.add_label(self.name)
        else:
            instance.__node__.remove_label(self.name)

    def __repr__(self):
        args = OrderedDict()
        if self.name is not None:
            args["name"] = self.name
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join("%s=%r" % arg for arg in args.items()))


def _resolve_class(model, current_module_name):
    if isinstance(model, type):
        return model
    module_name, _, class_name = model.rpartition(".")
    if not module_name:
        module_name = current_module_name
    module = __import__(module_name, fromlist=".")
    return getattr(module, class_name)


class Related(object):
    """ Descriptor for a set of related objects in a :class:`.Model`.

    Attributes:
        related_class: The class of object to which these relationships
                       connect. This class is used to coerce nodes to and
                       from :class:`Model` instances.
        relationship_type: The underlying relationship type for these
                           relationships. Note that the relationship
                           type should be unique for each class of related
                           object as the `related_class` is only used for
                           object coercion and not as part of the underlying
                           database query.
    """

    direction = UNDIRECTED

    def __init__(self, related_class, relationship_type=None):
        """ Initialise a property definition.

        Args:
            related_class: The class of object to which these relationships
                           connect.
            relationship_type: The underlying relationship type for these
                               relationships.
        """
        self.related_class = related_class
        self.relationship_type = relationship_type

    def __get__(self, instance, owner):
        cls = _resolve_class(self.related_class, instance.__class__.__module__)
        return instance.__ogm__.related(self.direction, self.relationship_type, cls)


class RelatedTo(Related):
    """ Descriptor for a set of related objects for a :class:`.Model`
    that are connected by outgoing relationships.
    """

    direction = OUTGOING


class RelatedFrom(Related):
    """ Descriptor for a set of related objects for a :class:`.Model`
    that are connected by incoming relationships.
    """

    direction = INCOMING


class RelatedObjects(object):
    """ A set of similarly-typed and similarly-related objects,
    relative to a central node.
    """

    def __init__(self, subject, node, direction, relationship_type, related_class):
        assert isinstance(direction, int) and not isinstance(direction, bool)
        self.subject = subject
        self.node = node
        self.relationship_type = relationship_type
        self.related_class = related_class
        self.__related_objects = None
        if direction > 0:
            self.__match_args = {"nodes": (self.node, None), "r_type": relationship_type}
            self.__start_node = False
            self.__end_node = True
            self.__relationship_pattern = "(a)-[_:%s]->(b)" % cypher_escape(relationship_type)
        elif direction < 0:
            self.__match_args = {"nodes": (None, self.node), "r_type": relationship_type}
            self.__start_node = True
            self.__end_node = False
            self.__relationship_pattern = "(a)<-[_:%s]-(b)" % cypher_escape(relationship_type)
        else:
            self.__match_args = {"nodes": {self.node, None}, "r_type": relationship_type}
            self.__start_node = True
            self.__end_node = True
            self.__relationship_pattern = "(a)-[_:%s]-(b)" % cypher_escape(relationship_type)

    def __iter__(self):
        for obj, _ in self._related_objects:
            yield obj

    def __len__(self):
        return len(self._related_objects)

    def __contains__(self, obj):
        if not isinstance(obj, Model):
            raise TypeError("Related objects must be Model instances")
        for related_object, _ in self._related_objects:
            if related_object == obj:
                return True
        return False

    @property
    def _related_objects(self):
        if self.__related_objects is None:
            self.__related_objects = []
            if self.node.graph:
                self.node.graph.update(lambda tx: self.__db_pull__(tx))
        return self.__related_objects

    def triples(self):
        """ Iterate through the related objects, yielding for each a
        triple of `(subject, (type, properties), object)`. This is the
        easiest way to see the full details of the underlying
        relationships.

            >>> from py2neo.ogm import *
            >>> from py2neo.ogm.models.movies import *
            >>> repo = Repository()
            >>> keanu = repo.match(Person).where(name="Keanu Reeves").first()
            >>> list(keanu.acted_in.triples())
            [(<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Julian Mercer']}), <Movie title="Something's Gotta Give">),
             (<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Shane Falco']}), <Movie title='The Replacements'>),
             (<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Johnny Mnemonic']}), <Movie title='Johnny Mnemonic'>),
             (<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Kevin Lomax']}), <Movie title="The Devil's Advocate">),
             (<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Neo']}), <Movie title='The Matrix Revolutions'>),
             (<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Neo']}), <Movie title='The Matrix Reloaded'>),
             (<Person name='Keanu Reeves'>, ('ACTED_IN', {'roles': ['Neo']}), <Movie title='The Matrix'>)]

        """
        for obj, properties in self._related_objects:
            yield self.subject, (self.relationship_type, properties), obj

    def add(self, obj, properties=None, **kwproperties):
        """ Add or update a related object.

        This method returns the number of new related objects that were
        added: 0 if an existing object was updated, or 1 if a new
        object was added.

        :param obj:
            :py:class:`.Model` to relate.
        :param properties:
            Dictionary of properties to attach to the relationship
            (optional).
        :param kwproperties:
            Additional keyword properties (optional).
        :returns:
            Number of new related objects that were added.
        """
        if not isinstance(obj, Model):
            raise TypeError("Related objects must be Model instances")
        related_objects = self._related_objects
        properties = dict(properties or {}, **kwproperties)
        for i, (related_object, p) in enumerate(related_objects):
            if related_object == obj:
                related_objects[i] = (obj, PropertyDict(p, **properties))
                return 0  # existing object was updated
        related_objects.append((obj, properties))
        return 1  # new object was added

    def clear(self):
        """ Remove all related objects from this set.
        """
        self._related_objects[:] = []

    def get(self, obj, key, default=None):
        """ Return a relationship property associated with a specific related object.

        :param obj: related object
        :param key: relationship property key
        :param default: default value, in case the key is not found
        :return: property value
        """
        if not isinstance(obj, Model):
            raise TypeError("Related objects must be Model instances")
        for related_object, properties in self._related_objects:
            if related_object == obj:
                return properties.get(key, default)
        return default

    def remove(self, obj):
        """ Remove a related object.

        :param obj: the :py:class:`.Model` to separate
        """
        if not isinstance(obj, Model):
            raise TypeError("Related objects must be Model instances")
        related_objects = self._related_objects
        related_objects[:] = [(related_object, properties)
                              for related_object, properties in related_objects
                              if related_object != obj]

    @deprecated("RelatedObjects.update is deprecated, "
                "please use RelatedObjects.add instead")
    def update(self, obj, properties=None, **kwproperties):
        """ Add or update a related object.

        :param obj: the :py:class:`.Model` to relate
        :param properties: dictionary of properties to attach to the relationship (optional)
        :param kwproperties: additional keyword properties (optional)
        """
        self.add(obj, properties, **kwproperties)

    def __db_pull__(self, tx):
        related_objects = {}
        for r in tx.graph.match(**self.__match_args):
            nodes = []
            n = self.node
            a = r.start_node
            b = r.end_node
            if a == b:
                nodes.append(a)
            else:
                if self.__start_node and a != n:
                    nodes.append(r.start_node)
                if self.__end_node and b != n:
                    nodes.append(r.end_node)
            for node in nodes:
                related_object = self.related_class.wrap(node)
                related_objects[node] = (related_object, PropertyDict(r))
        self._related_objects[:] = related_objects.values()

    def __db_push__(self, tx):
        related_objects = self._related_objects
        # 1. merge all nodes (create ones that don't)
        for related_object, _ in related_objects:
            tx.merge(related_object)
        # 2a. remove any relationships not in list of nodes
        subject_id = self.node.identity
        tx.run("MATCH %s WHERE id(a) = $x AND NOT id(b) IN $y DELETE _" % self.__relationship_pattern,
               x=subject_id, y=[obj.__node__.identity for obj, _ in related_objects])
        # 2b. merge all relationships
        for related_object, properties in related_objects:
            tx.run("MATCH (a) WHERE id(a) = $x MATCH (b) WHERE id(b) = $y "
                   "MERGE %s SET _ = $z" % self.__relationship_pattern,
                   x=subject_id, y=related_object.__node__.identity, z=properties)


class OGM(object):

    def __init__(self, subject, node):
        self.subject = subject
        self.node = node
        self._related = {}

    def all_related(self):
        """ Return an iterator through all :class:`.RelatedObjects`.
        """
        return iter(self._related.values())

    def related(self, direction, relationship_type, related_class):
        """ Return :class:`.RelatedObjects` for given criteria.
        """
        key = (direction, relationship_type)
        if key not in self._related:
            self._related[key] = RelatedObjects(self.subject, self.node, direction,
                                                relationship_type, related_class)
        return self._related[key]


class ModelType(type):

    def __new__(mcs, name, bases, attributes):
        for attr_name, attr in list(attributes.items()):
            if isinstance(attr, Property):
                if attr.key is None:
                    attr.key = attr_name
                if attr.__doc__ is attr.__class__.__doc__:
                    attr.__doc__ = repr(attr)
            elif isinstance(attr, Label):
                if attr.name is None:
                    attr.name = Words(attr_name).camel(upper_first=True)
                if attr.__doc__ is attr.__class__.__doc__:
                    attr.__doc__ = repr(attr)
            elif isinstance(attr, Related):
                if attr.relationship_type is None:
                    attr.relationship_type = Words(attr_name).upper("_")
                if attr.__doc__ is attr.__class__.__doc__:

                    def related_repr(obj):
                        try:
                            args = ":class:`%s`" % obj.related_class.__qualname__
                        except AttributeError:
                            args = ":class:`.%s`" % obj.related_class
                        if obj.relationship_type is not None:
                            args += ", relationship_type=%r" % obj.relationship_type
                        return "%s(%s)" % (obj.__class__.__name__, args)

                    attr.__doc__ = related_repr(attr)

        attributes.setdefault("__primarylabel__", name)

        primary_key = attributes.get("__primarykey__")
        if primary_key is None:
            for base in bases:
                if primary_key is None and hasattr(base, "__primarykey__"):
                    primary_key = getattr(base, "__primarykey__")
                    break
            else:
                primary_key = "__id__"
            attributes["__primarykey__"] = primary_key

        return super(ModelType, mcs).__new__(mcs, name, bases, attributes)


@metaclass(ModelType)
class Model(object):
    """ Base class for all OGM object classes.

    This class provides a default constructor, allowing initial
    property values to be provided. Positional arguments are mapped,
    in order, to primary key fields; keywords arguments are mapped
    directly by name.

    *Changed in 2020.0: this used to be called GraphObject, but was
    renamed to avoid ambiguity. The old name is still available as an
    alias.*

    *Changed in 2021.2: added default constructor.*
    """

    __primarylabel__ = None
    __primarykey__ = None

    __ogm = None

    def __init__(self, *values, **properties):
        p = {}
        if self.__primarykey__ == "__id__":
            primary_key = ()
        elif isinstance(self.__primarykey__, tuple):
            primary_key = self.__primarykey__
        else:
            primary_key = (self.__primarykey__,)
        for i, value in enumerate(values):
            try:
                key = primary_key[i]
            except IndexError:
                raise IndexError("Model constructor value at index %d does not correspond to "
                                 "any part of primary key %r" % (i, self.__primarykey__))
            else:
                p[key] = value
        p.update(properties)
        for key, value in p.items():
            try:
                setattr(self, key, value)
            except AttributeError:
                raise ValueError("No such property %r" % key)

    def __eq__(self, other):
        if self is other:
            return True
        try:
            self_node = self.__node__
            other_node = other.__node__
            if any(x is None for x in [self_node.graph, other_node.graph, self_node.identity, other_node.identity]):
                if self.__primarylabel__ != other.__primarylabel__:
                    return False
                if (self.__primarykey__ == other.__primarykey__ and
                        self.__primaryvalue__ == other.__primaryvalue__):
                    if self.__primarykey__ == "__id__" and self.__primaryvalue__ is None:
                        # If __id__ is the primary key but the value
                        # isn't yet set, assume the objects are
                        # not equal.
                        #
                        # See https://github.com/technige/py2neo/issues/839
                        #
                        return False
                    else:
                        return True
                else:
                    return False
            return (type(self) is type(other) and
                    self_node.graph == other_node.graph and
                    self_node.identity == other_node.identity)
        except (AttributeError, TypeError):
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def __ogm__(self):
        if self.__ogm is None:
            if isinstance(self.__primarylabel__, tuple):
                node = Node(*self.__primarylabel__)
            else:
                node = Node(self.__primarylabel__)
            node.__model__ = self.__class__
            self.__ogm = OGM(self, node)
        return self.__ogm

    @classmethod
    def wrap(cls, node):
        """ Convert a :class:`.Node` into a :class:`.Model`.

        :param node:
        :return:
        """
        if node is None:
            return None
        node.__model__ = cls
        obj = Model()
        obj.__ogm = OGM(obj, node)
        obj.__class__ = cls
        return obj

    @classmethod
    def match(cls, repository, primary_value=None):
        """ Select one or more nodes from the database, wrapped as instances of this class.

        :param repository: the :class:`.Repository` in which to match
        :param primary_value: value of the primary property (optional)
        :rtype: :class:`.ModelMatch`
        """
        return ModelMatcher(cls, repository).match(primary_value)

    def __repr__(self):
        return "<%s %s=%r>" % (self.__class__.__name__, self.__primarykey__, self.__primaryvalue__)

    @property
    def __primaryvalue__(self):
        node = self.__node__
        primary_key = self.__primarykey__
        if primary_key == "__id__":
            return node.identity
        elif isinstance(primary_key, tuple):
            return tuple(node[key] for key in primary_key)
        else:
            return node[primary_key]

    @property
    def __node__(self):
        """ The :class:`.Node` wrapped by this :class:`.Model`.
        """
        return self.__ogm__.node

    def __db_create__(self, tx):
        self.__db_merge__(tx)

    def __db_delete__(self, tx):
        ogm = self.__ogm__
        tx.delete(ogm.node)
        for related_objects in ogm.all_related():
            related_objects.clear()

    def __db_exists__(self, tx):
        return tx.exists(self.__node__)

    def __db_merge__(self, tx, primary_label=None, primary_key=None):
        ogm = self.__ogm__
        node = ogm.node
        if primary_label is None:
            primary_label = self.__primarylabel__
        if primary_key is None:
            primary_key = self.__primarykey__ or "__id__"
        if node.graph is None:
            if primary_key == "__id__":
                node.add_label(primary_label)
                tx.create(node)
            else:
                tx.merge(node, primary_label, primary_key)
            for related_objects in ogm.all_related():
                related_objects.__db_push__(tx)

    def __db_pull__(self, tx):
        ogm = self.__ogm__
        if ogm.node.graph is None:
            matcher = ModelMatcher(self.__class__, tx.graph)
            matcher._match_class = NodeMatch
            ogm.node = matcher.match(self.__primaryvalue__).first()
        tx.pull(ogm.node)
        for related_objects in ogm.all_related():
            related_objects.__db_pull__(tx)

    def __db_push__(self, tx):
        ogm = self.__ogm__
        node = ogm.node
        if node.graph is not None:
            tx.push(node)
        else:
            primary_key = self.__primarykey__ or "__id__"
            if primary_key == "__id__":
                tx.create(node)
            else:
                tx.merge(node)
        for related_objects in ogm.all_related():
            related_objects.__db_push__(tx)


# Alias for backward compatibility
GraphObject = Model


class ModelMatch(NodeMatch):
    """ A selection of :class:`.Model` instances that match a
    given set of criteria.
    """

    _object_class = Model

    def __iter__(self):
        """ Iterate through items drawn from the underlying repository
        that match the given criteria.
        """
        wrap = self._object_class.wrap
        for node in super(ModelMatch, self).__iter__():
            yield wrap(node)

    def first(self):
        """ Return the first item that matches the given criteria.
        """
        return self._object_class.wrap(super(ModelMatch, self).first())


class ModelMatcher(NodeMatcher):

    _match_class = ModelMatch

    @classmethod
    def _coerce_to_graph(cls, obj):
        if isinstance(obj, Repository):
            return obj.graph
        elif isinstance(obj, Graph):
            return obj
        else:
            raise TypeError("Cannot coerce object %r to Graph" % obj)

    def __init__(self, object_class, repository):
        NodeMatcher.__init__(self, self._coerce_to_graph(repository))
        self._object_class = object_class
        self._match_class = type("%sMatch" % self._object_class.__name__,
                                 (ModelMatch,), {"_object_class": object_class})

    def match(self, primary_value=None):
        cls = self._object_class
        if cls.__primarykey__ == "__id__":
            match = NodeMatcher.match(self, cls.__primarylabel__)
            if primary_value is not None:
                match = match.where("id(_) = %d" % primary_value)
        elif primary_value is None:
            match = NodeMatcher.match(self, cls.__primarylabel__)
        else:
            match = NodeMatcher.match(self, cls.__primarylabel__).where(**{cls.__primarykey__: primary_value})
        return match


class Repository(object):
    """ Storage container for :class:`.Model` instances.

    The constructor for this class has an identical signature to that
    for the :class:`.Graph` class. For example::

        >>> from py2neo.ogm import Repository
        >>> from py2neo.ogm.models.movies import Movie
        >>> repo = Repository("bolt://neo4j@localhost:7687", password="password")
        >>> repo.match(Movie, "The Matrix").first()
        <Movie title='The Matrix'>

    *New in version 2020.0. In earlier versions, a :class:`.Graph` was
    required to co-ordinate all reads and writes to the remote
    database. This class completely replaces that, removing the need
    to import from any other packages when using OGM.*
    """

    @classmethod
    def wrap(cls, graph):
        """ Wrap an existing :class:`.Graph` object as a
        :class:`.Repository`.
        """
        obj = object.__new__(Repository)
        obj.graph = graph
        return obj

    def __init__(self, profile=None, name=None, **settings):
        self.graph = Graph(profile, name=name, **settings)

    def __repr__(self):
        return "<Repository profile=%r>" % (self.graph.service.profile,)

    def reload(self, obj):
        """ Reload data from the remote graph into the local object.
        """
        self.graph.pull(obj)

    def save(self, *objects):
        """ Save data from the local object into the remote graph.

        Each object in `objects` can be a :class:`Model` object or an
        iterable of such objects.

        :param objects: :class:`Model` objects to save.
        """

        def push_all(tx, iterable):
            for obj in iterable:
                if hasattr(obj, "__db_push__"):
                    tx.push(obj)
                elif hasattr(obj, "__iter__"):
                    push_all(tx, obj)
                else:
                    raise ValueError("Object %r is neither savable "
                                     "nor iterable" % obj)

        self.graph.update(lambda tx: push_all(tx, objects))

    def delete(self, obj):
        """ Delete the object in the remote graph.
        """
        self.graph.delete(obj)

    def exists(self, obj):
        """ Check whether the object exists in the remote graph.
        """
        return self.graph.exists(obj)

    def match(self, model, primary_value=None):
        """ Select one or more objects from the remote graph.

        :param model: the :class:`.Model` subclass to match
        :param primary_value: value of the primary property (optional)
        :rtype: :class:`.ModelMatch`
        """
        return ModelMatcher(model, self).match(primary_value)

    def get(self, model, primary_value=None):
        """ Match and return a single object from the remote graph.

        :param model: the :class:`.Model` subclass to match
        :param primary_value: value of the primary property (optional)
        :rtype: :class:`.Model`
        """
        return self.match(model, primary_value).first()

    @deprecated("Repository.create is a compatibility alias, "
                "please use Repository.save instead")
    def create(self, obj):
        self.graph.create(obj)

    @deprecated("Repository.merge is a compatibility alias, "
                "please use Repository.save instead")
    def merge(self, obj):
        self.graph.merge(obj)

    @deprecated("Repository.pull is a compatibility alias, "
                "please use Repository.load instead")
    def pull(self, obj):
        self.graph.pull(obj)

    @deprecated("Repository.push is a compatibility alias, "
                "please use Repository.save instead")
    def push(self, obj):
        self.graph.push(obj)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

from collections import OrderedDict
from logging import getLogger
from json import dumps as json_dumps, loads as json_loads

from packaging.version import Version
from six import raise_from
from urllib3 import HTTPConnectionPool, HTTPSConnectionPool, make_headers
from urllib3.exceptions import HTTPError

from py2neo import ConnectionProfile
from py2neo.compat import urlsplit
from py2neo.client import http_user_agent, Connection, TransactionRef, Result, Bookmark
from py2neo.client.json import JSONHydrant, dehydrate
from py2neo.errors import Neo4jError, ConnectionUnavailable, ProtocolError


log = getLogger(__name__)


class HTTP(Connection):

    @classmethod
    def default_hydrant(cls, profile, graph):
        return JSONHydrant(graph)

    @classmethod
    def open(cls, profile=None, user_agent=None, on_release=None, on_broken=None):
        """ Open an HTTP connection to a server.

        :param profile: :class:`.ConnectionProfile` detailing how and
            where to connect
        :param user_agent:
        :param on_release:
        :param on_broken:
        :returns: :class:`.HTTP` connection object
        :raises: :class:`.ConnectionUnavailable` if a connection cannot
            be opened
        """
        if profile is None:
            profile = ConnectionProfile(scheme="http")
        try:
            http = cls(profile, on_release=on_release)
            http._hello(user_agent or http_user_agent())
            return http
        except HTTPError as error:
            raise_from(ConnectionUnavailable("Cannot open connection to %r", profile), error)

    def __init__(self, profile, on_release=None):
        super(HTTP, self).__init__(profile, on_release=on_release)
        self.http_pool = None
        self.headers = {}
        self.__closed = False
        self._make_pool(profile)

    def _make_pool(self, profile):
        if profile.secure:
            from ssl import CERT_NONE, CERT_REQUIRED
            from certifi import where as cert_where
            self.http_pool = HTTPSConnectionPool(
                host=profile.host,
                port=profile.port_number,
                maxsize=1,
                block=True,
                cert_reqs=CERT_REQUIRED if profile.verify else CERT_NONE,
                ca_certs=cert_where()
            )
        else:
            self.http_pool = HTTPConnectionPool(
                host=profile.host,
                port=profile.port_number,
                maxsize=1,
                block=True,
            )

    def close(self):
        self.http_pool.close()
        self.__closed = True

    @property
    def closed(self):
        return self.__closed

    @property
    def broken(self):
        return False

    def _hello(self, user_agent):
        self.headers.update(make_headers(basic_auth=":".join(self.profile.auth),
                                         user_agent=user_agent))
        r = self.http_pool.request(method="GET",
                                   url="/",
                                   headers=dict(self.headers))
        metadata = json_loads(r.data.decode("utf-8"))
        if "neo4j_version" in metadata:
            # {
            #   "bolt_routing" : "neo4j://localhost:7687",
            #   "transaction" : "http://localhost:7474/db/{databaseName}/tx",
            #   "bolt_direct" : "bolt://localhost:7687",
            #   "neo4j_version" : "4.0.0",
            #   "neo4j_edition" : "community"
            # }
            self._neo4j_version = Version(metadata["neo4j_version"])  # Neo4j 4.x
            self._neo4j_edition = metadata["neo4j_edition"]
        else:                               # Neo4j 3.x
            # {
            #   "data" : "http://localhost:7474/db/data/",
            #   "management" : "http://localhost:7474/db/manage/",
            #   "bolt" : "bolt://localhost:7687"
            # }
            r = self.http_pool.request(method="GET",
                                       url="/db/data/",
                                       headers=dict(self.headers))
            metadata = json_loads(r.data.decode("utf-8"))
            # {
            #   "extensions" : { },
            #   "node" : "http://localhost:7474/db/data/node",
            #   "relationship" : "http://localhost:7474/db/data/relationship",
            #   "node_index" : "http://localhost:7474/db/data/index/node",
            #   "relationship_index" : "http://localhost:7474/db/data/index/relationship",
            #   "extensions_info" : "http://localhost:7474/db/data/ext",
            #   "relationship_types" : "http://localhost:7474/db/data/relationship/types",
            #   "batch" : "http://localhost:7474/db/data/batch",
            #   "cypher" : "http://localhost:7474/db/data/cypher",
            #   "indexes" : "http://localhost:7474/db/data/schema/index",
            #   "constraints" : "http://localhost:7474/db/data/schema/constraint",
            #   "transaction" : "http://localhost:7474/db/data/transaction",
            #   "node_labels" : "http://localhost:7474/db/data/labels",
            #   "neo4j_version" : "3.5.12"
            # }
            self._neo4j_version = Version(metadata["neo4j_version"])  # Neo4j 3.x
        self.server_agent = "Neo4j/{}".format(self._neo4j_version)

        # Given the root discovery endpoint isn't authenticated, we don't
        # catch incorrect passwords here, and this wouldn't then be signalled
        # to the user until later on. So here, we make a second call to a
        # different URL for that reason only.
        r = self.http_pool.request(method="GET",
                                   url="/db/data/",
                                   headers=dict(self.headers))
        data = r.data.decode("utf-8")
        rs = HTTPResponse.from_json(r.status, data or "{}")
        rs.audit()

    def fast_forward(self, bookmark):
        raise NotImplementedError("Bookmarking is not yet supported over HTTP")

    def auto_run(self, cypher, parameters=None, graph_name=None, readonly=False,
                 # after=None, metadata=None, timeout=None
                 ):
        try:
            if graph_name and not self.supports_multi():
                raise TypeError("Neo4j {} does not support "
                                "named graphs".format(self.neo4j_version))
            # if readonly:
            #     log.warning("Readonly transactions are not supported over HTTP")
            r = self._post(HTTPTransactionRef.autocommit_uri(graph_name), cypher, parameters)
            rs = HTTPResponse.from_json(r.status, r.data.decode("utf-8"))
            rs.audit()
            return HTTPResult(HTTPTransactionRef(graph_name), rs.result(), profile=self.profile)
        finally:
            self.release()

    def begin(self, graph_name, readonly=False,
              # after=None, metadata=None, timeout=None
              ):
        if graph_name and not self.supports_multi():
            raise TypeError("Neo4j {} does not support "
                            "named graphs".format(self.neo4j_version))
        try:
            # if readonly:
            #     log.warning("Readonly transactions are not supported over HTTP")
            # if after:
            #     raise TypeError("Bookmarks are not supported over HTTP")
            # if metadata:
            #     raise TypeError("Transaction metadata is not supported over HTTP")
            # if timeout:
            #     raise TypeError("Transaction timeouts are not supported over HTTP")
            r = self._post(HTTPTransactionRef.begin_uri(graph_name))
            rs = HTTPResponse.from_json(r.status, r.data.decode("utf-8"))
            location_path = urlsplit(r.headers["Location"]).path
            tx = HTTPTransactionRef(graph_name, location_path.rpartition("/")[-1])
            rs.audit(tx)
            return tx
        finally:
            self.release()

    def commit(self, tx):
        try:
            if tx.broken:
                raise ValueError("Transaction is broken")
            r = self._post(tx.commit_uri())
        except ProtocolError:
            tx.mark_broken()
            raise
        else:
            rs = HTTPResponse.from_json(r.status, r.data.decode("utf-8"))
            rs.audit(tx)
            return Bookmark()
        finally:
            self.release()

    def rollback(self, tx):
        try:
            if tx.broken:
                raise ValueError("Transaction is broken")
            r = self._delete(tx.uri())
        except ProtocolError:
            tx.mark_broken()
            raise
        else:
            rs = HTTPResponse.from_json(r.status, r.data.decode("utf-8"))
            rs.audit(tx)
            return Bookmark()
        finally:
            self.release()

    def run(self, tx, cypher, parameters=None):
        try:
            r = self._post(tx.uri(), cypher, parameters)
        except ProtocolError:
            tx.mark_broken()
            raise
        else:
            rs = HTTPResponse.from_json(r.status, r.data.decode("utf-8"))
            rs.audit(tx)
            return HTTPResult(tx, rs.result(), profile=self.profile)
        finally:
            self.release()

    def pull(self, result, n=-1):
        try:
            # TODO: lower-memory algorithm that doesn't require
            #  duplication of potentially large data sets
            if n == -1:
                result._buffer.extend(result._data)
                result._data[:] = []
            else:
                result._buffer.extend(result._data[:n])
                result._data[:n] = []
        finally:
            self.release()

    def discard(self, result):
        try:
            result._data[:] = []
        finally:
            self.release()

    def _get_http_profiles(self):
        scheme = "https" if self.profile.secure else "http"
        profiles = {}
        try:
            result = self.auto_run("CALL dbms.cluster.overview")
            self.pull(result)
            while True:
                record = result.take()
                if record is None:
                    break
                key = None
                value = None
                for address in record[1]:
                    profile = ConnectionProfile(address)
                    if profile.scheme == "bolt":
                        key = profile.address
                    elif profile.scheme == scheme:
                        value = profile.address
                    else:
                        pass  # unusable value (should never happen)
                if key and value:
                    profiles[key] = value
        except Neo4jError as error:
            if error.title == "ProcedureNotFound":
                raise_from(TypeError("Neo4j service does not expose a cluster overview"), error)
            else:
                raise
        else:
            return profiles, scheme

    def route(self, graph_name=None, context=None):
        try:
            if self._neo4j_version >= Version("4.0"):
                routers, readers, writers, ttl = self._route4(graph_name, context)
            else:
                routers, readers, writers, ttl = self._route1(graph_name, context)
            profiles, scheme = self._get_http_profiles()  # Convert Bolt addresses to HTTP
            return ([ConnectionProfile(_, scheme=scheme, address=profiles[_.address]) for _ in routers],
                    [ConnectionProfile(_, scheme=scheme, address=profiles[_.address]) for _ in readers],
                    [ConnectionProfile(_, scheme=scheme, address=profiles[_.address]) for _ in writers],
                ttl)
        finally:
            self.release()

    def sync(self, result):
        pass

    def fetch(self, result):
        record = result.take()
        return record

    def _post(self, url, statement=None, parameters=None):
        log.debug("POST %r %r %r", url, statement, parameters)
        if statement:
            statements = [
                OrderedDict([
                    ("statement", statement),
                    ("parameters", dehydrate(parameters or {})),
                    ("resultDataContents", ["REST"]),
                    ("includeStats", True),
                ])
            ]
        else:
            statements = []
        try:
            return self.http_pool.request(method="POST",
                                          url=url,
                                          headers=dict(self.headers, **{"Content-Type": "application/json"}),
                                          body=json_dumps({"statements": statements}))
        except HTTPError as error:
            raise_from(ProtocolError("Failed to POST to %r" % url), error)

    def _delete(self, url):
        log.debug("DELETE %r", url)
        try:
            return self.http_pool.request(method="DELETE",
                                          url=url,
                                          headers=dict(self.headers))
        except HTTPError as error:
            raise_from(ProtocolError("Failed to DELETE %r" % url), error)

    def supports_multi(self):
        return self._neo4j_version >= Version("4.0")


class HTTPTransactionRef(TransactionRef):

    def __init__(self, graph_name, txid=None, readonly=False):
        super(HTTPTransactionRef, self).__init__(graph_name, txid, readonly)
        self.failure = None

    def __bool__(self):
        return not self.broken

    __nonzero__ = __bool__

    @classmethod
    def autocommit_uri(cls, graph_name):
        if graph_name:
            return "/db/{}/tx/commit".format(graph_name)
        else:
            return "/db/data/transaction/commit"

    @classmethod
    def begin_uri(cls, graph_name):
        if graph_name:
            return "/db/{}/tx".format(graph_name)
        else:
            return "/db/data/transaction"

    def uri(self):
        if self.graph_name:
            return "/db/{}/tx/{}".format(self.graph_name, self.txid)
        else:
            return "/db/data/transaction/{}".format(self.txid)

    def commit_uri(self):
        if self.graph_name:
            return "/db/{}/tx/{}/commit".format(self.graph_name, self.txid)
        else:
            return "/db/data/transaction/{}/commit".format(self.txid)


class HTTPResult(Result):

    def __init__(self, tx, result, profile):
        Result.__init__(self, tx)
        self._profile = profile
        self._columns = result.get("columns", ())
        self._data = result.get("data", [])
        self._buffer = []
        self._summary = {}
        if "stats" in result:
            self._summary["stats"] = result["stats"]
        self._cursor = 0

    @property
    def offline(self):
        return True

    @property
    def profile(self):
        return self._profile

    def fields(self):
        return self._columns

    def summary(self):
        return self._summary

    def take(self):
        try:
            record = self._buffer[self._cursor]["rest"]
        except IndexError:
            return None
        else:
            self._cursor += 1
            return record

    def peek(self, limit):
        records = []
        for i in range(limit):
            try:
                records.append(self._buffer[self._cursor + i]["rest"])
            except IndexError:
                break
        return records


class HTTPResponse(object):

    @classmethod
    def from_json(cls, status, data):
        try:
            content = json_loads(data, object_hook=JSONHydrant.json_to_packstream)
        except ValueError as error:
            raise_from(ProtocolError("Cannot decode response content as JSON"), error)
        else:
            return cls(status, content)

    def __init__(self, status, content):
        self._status = status
        self._content = content

    @property
    def status(self):
        return self._status

    def columns(self):
        return tuple(self._content.get("columns", ()))

    def result(self, index=0):
        try:
            results = self._content["results"]
        except KeyError:
            return {}  # TODO return None, somehow
        else:
            try:
                return results[index]
            except IndexError:
                return {}  # TODO return None, somehow

    def stats(self):
        return self._content.get("stats", {})

    def errors(self):
        return self._content.get("errors", [])

    def audit(self, tx=None):
        if self.errors():
            failure = Neo4jError.hydrate(self.errors().pop(0))
            if tx is not None:
                tx.mark_broken()
            raise failure



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright 2011-2021, Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import, print_function


__all__ = [
    "GraphService",
    "Graph",
    "SystemGraph",
    "Schema",
    "Transaction",
]


from inspect import isgenerator
from time import sleep

from py2neo.compat import (deprecated,
                           Sequence,
                           Mapping)
from py2neo.cypher import Cursor, cypher_escape
from py2neo.cypher.proc import ProcedureLibrary
from py2neo.errors import (Neo4jError,
                           ConnectionUnavailable,
                           ConnectionBroken,
                           ConnectionLimit,
                           ServiceUnavailable,
                           WriteServiceUnavailable)
from py2neo.matching import NodeMatcher, RelationshipMatcher


class GraphService(object):
    """ The :class:`.GraphService` class is the top-level accessor for
    an entire Neo4j graph database management system (DBMS). Within the
    py2neo object hierarchy, a :class:`.GraphService` contains one or
    more :class:`.Graph` objects in which data storage and retrieval
    activity chiefly occurs.

    An explicit URI can be passed to the constructor::

        >>> from py2neo import GraphService
        >>> gs = GraphService("bolt://camelot.example.com:7687")

    Alternatively, the default value of ``bolt://localhost:7687`` is
    used::

        >>> default_gs = GraphService()
        >>> default_gs
        <GraphService uri='bolt://localhost:7687'>

    .. note::

        Some attributes of this class available in earlier versions of
        py2neo are no longer available, specifically
        ``kernel_start_time``, ``primitive_counts``,
        ``store_creation_time``, ``store_file_sizes`` and ``store_id``,
        along with the ``query_jmx`` method. This is due to a change in
        Neo4j 4.0 relating to how certain system metadata is exposed.
        Replacement functionality may be reintroduced in a future
        py2neo release.

    *Changed in 2020.0: this class was formerly known as 'Database',
    but was renamed to avoid confusion with the concept of the same
    name introduced with the multi-database feature of Neo4j 4.0.*

    .. describe:: iter(graph_service)

        Yield all named graphs.

        For Neo4j 4.0 and above, this yields the names returned by a
        ``SHOW DATABASES`` query. For earlier versions, this yields no
        entries, since the one and only graph in these versions is not
        named.

        *New in version 2020.0.*

    .. describe:: graph_service[name]

        Access a :class:`.Graph` by name.

        *New in version 2020.0.*

    """

    _connector = None

    _graphs = None

    def __init__(self, profile=None, **settings):
        from py2neo import ServiceProfile
        from py2neo.client import Connector
        connector_settings = {
            "user_agent": settings.pop("user_agent", None),
            "init_size": settings.pop("init_size", None),
            "max_size": settings.pop("max_size", None),
            "max_age": settings.pop("max_age", None),
            "routing_refresh_ttl": settings.pop("routing_refresh_ttl", None),
        }
        profile = ServiceProfile(profile, **settings)
        if connector_settings["init_size"] is None and not profile.routing:
            # Ensures credentials are checked on construction
            connector_settings["init_size"] = 1
        self._connector = Connector(profile, **connector_settings)
        self._graphs = {}

    def __repr__(self):
        class_name = self.__class__.__name__
        profile = self._connector.profile
        return "<%s uri=%r secure=%r user_agent=%r>" % (
            class_name, profile.uri, profile.secure, self._connector.user_agent)

    def __eq__(self, other):
        try:
            return self.uri == other.uri
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self._connector)

    def __getitem__(self, graph_name):
        if graph_name is None:
            graph_name = self._connector.default_graph_name()
        elif graph_name not in self._connector.graph_names():
            raise KeyError("Graph {!r} does not exist for "
                           "service {!r}".format(graph_name, self._connector.profile.uri))
        if graph_name not in self._graphs:
            graph_class = SystemGraph if graph_name == "system" else Graph
            self._graphs[graph_name] = graph_class(self.profile, name=graph_name)
        return self._graphs[graph_name]

    def __iter__(self):
        return iter(self._connector.graph_names())

    @property
    def connector(self):
        """ The :class:`.Connector` providing communication for this
        graph service.

        *New in version 2020.0.*
        """
        return self._connector

    @property
    def profile(self):
        """ The :class:`.ConnectionProfile` for which this graph
        service is configured. This attribute is simply a shortcut
        for ``connector.profile``.

        *New in version 2020.0.*
        """
        return self.connector.profile

    @property
    def uri(self):
        """ The URI to which this graph service is connected. This
        attribute is simply a shortcut for ``connector.profile.uri``.
        """
        return self.profile.uri

    @property
    def default_graph(self):
        """ The default :class:`.Graph` exposed by this graph service.
        """
        return self[None]

    @property
    def system_graph(self):
        """ The :class:`.SystemGraph` exposed by this graph service.

        *New in version 2020.0.*
        """
        return self["system"]

    def keys(self):
        """ Return a list of all :class:`.Graph` names exposed by this
        graph service.

        *New in version 2020.0.*
        """
        return list(self)

    @property
    def kernel_version(self):
        """ The :class:`~packaging.version.Version` of Neo4j running.
        """
        from packaging.version import Version
        components = self.default_graph.call("dbms.components").data()
        kernel_component = [component for component in components
                            if component["name"] == "Neo4j Kernel"][0]
        version_string = kernel_component["versions"][0]
        return Version(version_string)

    @property
    def product(self):
        """ The product name.
        """
        record = next(self.default_graph.call("dbms.components"))
        return "%s %s (%s)" % (record[0], " ".join(record[1]), record[2].title())

    @property
    def config(self):
        """ A dictionary of the configuration parameters used to
        configure Neo4j.

            >>> gs.config['dbms.connectors.default_advertised_address']
            'localhost'

        """
        return {record["name"]: record["value"]
                for record in self.default_graph.call("dbms.listConfig")}


class Graph(object):
    """ The `Graph` class provides a handle to an individual named
    graph database exposed by a Neo4j graph database service.

    Connection details are provided using either a URI or a
    :class:`.ConnectionProfile`, plus individual settings, if required.

    The `name` argument allows selection of a graph database by name.
    When working with Neo4j 4.0 and above, this can be any name defined
    in the system catalogue, a full list of which can be obtained
    through the Cypher ``SHOW DATABASES`` command. Passing `None` here
    will select the default database, as defined on the server. For
    earlier versions of Neo4j, the `name` must be set to `None`.

        >>> from py2neo import Graph
        >>> sales = Graph("bolt+s://g.example.com:7687", name="sales")
        >>> sales.run("MATCH (c:Customer) RETURN c.name")
         c.name
        ---------------
         John Smith
         Amy Pond
         Rory Williams

    The `system graph`, which is available in all 4.x+ product editions,
    can also be accessed via the :class:`.SystemGraph` class.

        >>> from py2neo import SystemGraph
        >>> sg = SystemGraph("bolt+s://g.example.com:7687")
        >>> sg.call("dbms.security.listUsers")
         username | roles | flags
        ----------|-------|-------
         neo4j    |  null | []

    In addition to the core `connection details <#getting-connected>`_
    that can be passed to the constructor, the :class:`.Graph` class
    can accept several other settings:

    ===================  ========================================================  ==============  =========================
    Keyword              Description                                               Type            Default
    ===================  ========================================================  ==============  =========================
    ``user_agent``       User agent to send for all connections                    str             `(depends on URI scheme)`
    ``max_connections``  The maximum number of simultaneous connections permitted  int             40
    ===================  ========================================================  ==============  =========================

    Once obtained, the `Graph` instance provides direct or indirect
    access to most of the functionality available within py2neo.
    """

    #: The :class:`.GraphService` to which this :class:`.Graph` belongs.
    service = None

    #: The :class:`.Schema` resource for this :class:`.Graph`.
    schema = None

    def __init__(self, profile=None, name=None, **settings):
        self.service = GraphService(profile, **settings)
        self.__name__ = name
        self.schema = Schema(self)
        self._procedures = ProcedureLibrary(self)

    def __repr__(self):
        if self.name is None:
            return "%s(%r)" % (self.__class__.__name__, self.service.uri)
        else:
            return "%s(%r, name=%r)" % (self.__class__.__name__, self.service.uri, self.name)

    def __eq__(self, other):
        try:
            return self.service == other.service and self.__name__ == other.__name__
        except (AttributeError, TypeError):
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.relationships)

    def __bool__(self):
        return True

    __nonzero__ = __bool__

    @property
    def name(self):
        """ The name of this graph.

        *New in version 2020.0.*
        """
        return self.__name__

    # TRANSACTION MANAGEMENT #

    def auto(self, readonly=False,
             # after=None, metadata=None, timeout=None
             ):
        """ Create a new auto-commit :class:`~py2neo.Transaction`.

        :param readonly: if :py:const:`True`, will begin a readonly
            transaction, otherwise will begin as read-write

        *New in version 2020.0.*
        """
        return Transaction(self, autocommit=True, readonly=readonly,
                           # after, metadata, timeout
                           )

    def begin(self, readonly=False,
              # after=None, metadata=None, timeout=None
              ):
        """ Begin a new :class:`~py2neo.Transaction`.

        :param readonly: if :py:const:`True`, will begin a readonly
            transaction, otherwise will begin as read-write

        *Changed in version 2021.1: the 'autocommit' argument has been
        removed. Use the 'auto' method instead.*
        """
        return Transaction(self, autocommit=False, readonly=readonly,
                           # after, metadata, timeout
                           )

    def commit(self, tx):
        """ Commit a transaction.

        *New in version 2021.1.*
        """
        if tx is None:
            return
        if not isinstance(tx, Transaction):
            raise TypeError("Bad transaction %r" % tx)
        if tx.closed:
            raise TypeError("Cannot commit closed transaction")
        try:
            summary = self.service.connector.commit(tx.ref)
            tx._bookmark = summary["bookmark"]
            tx._profile = summary["profile"]
            tx._time = summary["time"]
        finally:
            tx._closed = True

    def rollback(self, tx):
        """ Rollback a transaction.

        *New in version 2021.1.*
        """
        if tx is None or tx.closed:
            return
        if not isinstance(tx, Transaction):
            raise TypeError("Bad transaction %r" % tx)
        try:
            summary = self.service.connector.rollback(tx.ref)
            tx._bookmark = summary["bookmark"]
            tx._profile = summary["profile"]
            tx._time = summary["time"]
        except (ConnectionUnavailable, ConnectionBroken):
            pass
        finally:
            tx._closed = True

    # CYPHER EXECUTION #

    def run(self, cypher, parameters=None, **kwparameters):
        """ Run a single read/write query within an auto-commit
        :class:`~py2neo.Transaction`.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :param kwparameters: extra parameters supplied as keyword
            arguments
        :return:
        """
        return self.auto().run(cypher, parameters, **kwparameters)

    def evaluate(self, cypher, parameters=None, **kwparameters):
        """ Run a :meth:`~py2neo.Transaction.evaluate` operation within an
        auto-commit :class:`~py2neo.Transaction`.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :return: first value from the first record returned or
                 :py:const:`None`.
        """
        return self.run(cypher, parameters, **kwparameters).evaluate()

    def update(self, cypher, parameters=None, timeout=None):
        """ Execute a transactional unit of work that carries out
        write operations but does not return a result.

        The `cypher` argument can be either a single string statement
        or a callable that accepts a :class:`~py2neo.Transaction`
        object as its first argument. Additional arguments can be
        passed though the `args` and `kwargs` arguments of this method.

        The unit of work may be called multiple times if earlier
        attempts fail due to connectivity or other transient errors.
        As such, the function should have no non-idempotent side
        effects.

        Note that hybrid read-write queries (i.e. those which carry out
        updates *and* return a result should use the non-retrying
        :class:`.run` method instead.

        :param cypher: cypher string or transaction function containing
            a unit of work
        :param parameters: cypher parameter map or function arguments
        :param timeout:
        :raises WriteServiceUnavailable: if the update does not
            successfully complete
        """
        if callable(cypher):
            if parameters is None:
                self._update(cypher, timeout=timeout)
            elif (isinstance(parameters, tuple) and len(parameters) == 2 and
                    isinstance(parameters[0], Sequence) and isinstance(parameters[1], Mapping)):
                self._update(lambda tx: cypher(tx, *parameters[0], **parameters[1]),
                             timeout=timeout)
            elif isinstance(parameters, Sequence):
                self._update(lambda tx: cypher(tx, *parameters), timeout=timeout)
            elif isinstance(parameters, Mapping):
                self._update(lambda tx: cypher(tx, **parameters), timeout=timeout)
            else:
                raise TypeError("Unrecognised parameter type")
        else:
            self._update(lambda tx: tx.update(cypher, parameters), timeout=timeout)

    def _update(self, f, timeout=None):
        from py2neo.timing import Timer
        # TODO: logging
        n = 0
        for _ in Timer.repeat(at_least=3, timeout=timeout):
            n += 1
            tx = None
            try:
                tx = self.begin(
                                # after=after, metadata=metadata, timeout=timeout
                                )
                value = f(tx)
                if isgenerator(value):
                    _ = list(value)     # exhaust the generator
                self.commit(tx)
            except (ConnectionUnavailable, ConnectionBroken, ConnectionLimit):
                self.rollback(tx)
                continue
            except Neo4jError as error:
                self.rollback(tx)
                if error.should_invalidate_routing_table():
                    self.service.connector.invalidate_routing_table(self.name)
                if error.should_retry():
                    continue
                else:
                    raise
            except Exception:
                self.rollback(tx)
                raise
            else:
                return
        raise WriteServiceUnavailable("Failed to execute update after %r tries" % n)

    def query(self, cypher, parameters=None, timeout=None):
        """ Run a single query within a readonly auto-commit
        :class:`~py2neo.Transaction`.

        Should query execution fail due to connection failure or other
        transient error, retries will be attempted. There will be a
        minimum of three attempts, and retries will continue until the
        timeout passes.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :param timeout:
        :returns: result :class:`~.cypher.Cursor` object
        :raises TypeError: if the underlying connection profile does not
            support readonly transactions
        :raises ServiceUnavailable: if the query does not successfully
            complete

        *Refactored from read to query in version 2021.1*
        """
        from py2neo.timing import Timer
        # TODO: logging
        n = 0
        for _ in Timer.repeat(at_least=3, timeout=timeout):
            n += 1
            try:
                result = self.auto(readonly=True).run(cypher, parameters)
            except (ConnectionUnavailable, ConnectionBroken, ConnectionLimit):
                continue
            except Neo4jError as error:
                if error.should_invalidate_routing_table():
                    self.service.connector.invalidate_routing_table(self.name)
                if error.should_retry():
                    continue
                else:
                    raise
            else:
                return result
        raise ServiceUnavailable("Failed to execute query after %r tries" % n)

    @property
    def call(self):
        """ Accessor for listing and calling procedures.

        This property contains a :class:`.ProcedureLibrary` object tied
        to this graph, which provides links to Cypher procedures in
        the underlying implementation.

        Calling a procedure requires only the regular Python function
        call syntax::

            >>> g = Graph()
            >>> g.call.dbms.components()
             name         | versions   | edition
            --------------|------------|-----------
             Neo4j Kernel | ['3.5.12'] | community

        The object returned from the call is a
        :class:`~py2neo.Cursor` object, identical to
        that obtained from running a normal Cypher query, and can
        therefore be consumed in a similar way.

        Procedure names can alternatively be supplied as a string::

            >>> g.call["dbms.components"]()
             name         | versions   | edition
            --------------|------------|-----------
             Neo4j Kernel | ['3.5.12'] | community

        Using :func:`dir` or :func:`iter` on the `call` attribute will
        yield a list of available procedure names.

        *New in version 2020.0.*
        """
        return self._procedures

    def delete_all(self):
        """ Delete all nodes and relationships from this :class:`.Graph`.

        .. warning::
            This method will permanently remove **all** nodes and relationships
            from the graph and cannot be undone.
        """
        self.run("MATCH (a) DETACH DELETE a")

    @deprecated("The graph.read(...) method is deprecated, "
                "use graph.query(...) instead")
    def read(self, cypher, parameters=None, **kwparameters):
        return self.query(cypher, dict(parameters or {}, **kwparameters))

    # SUBGRAPH OPERATIONS #

    def create(self, subgraph):
        """ Run a :meth:`~py2neo.Transaction.create` operation within a
        :class:`~py2neo.Transaction`.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        self.update(lambda tx: tx.create(subgraph))

    def delete(self, subgraph):
        """ Run a :meth:`~py2neo.Transaction.delete` operation within an
        auto-commit :class:`~py2neo.Transaction`. To delete only the
        relationships, use the :meth:`.separate` method.

        Note that only entities which are bound to corresponding
        remote entities though the ``graph`` and ``identity``
        attributes will trigger a deletion.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        """
        self.update(lambda tx: tx.delete(subgraph))

    def exists(self, subgraph):
        """ Run a :meth:`~py2neo.Transaction.exists` operation within an
        auto-commit :class:`~py2neo.Transaction`.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        :return:
        """
        return self.auto(readonly=True).exists(subgraph)

    def match(self, nodes=None, r_type=None, limit=None):
        """ Match and return all relationships with specific criteria.

        For example, to find all of Alice's friends::

            for rel in graph.match((alice, ), r_type="FRIEND"):
                print(rel.end_node["name"])

        :param nodes: Sequence or Set of start and end nodes (:const:`None` means any node);
                a Set implies a match in any direction
        :param r_type: type of relationships to match (:const:`None` means any type)
        :param limit: maximum number of relationships to match (:const:`None` means unlimited)
        """
        return RelationshipMatcher(self).match(nodes=nodes, r_type=r_type).limit(limit)

    def match_one(self, nodes=None, r_type=None):
        """ Match and return one relationship with specific criteria.

        :param nodes: Sequence or Set of start and end nodes (:const:`None` means any node);
                a Set implies a match in any direction
        :param r_type: type of relationships to match (:const:`None` means any type)
        """
        matches = self.match(nodes=nodes, r_type=r_type, limit=1)
        rels = list(matches)
        if rels:
            return rels[0]
        else:
            return None

    def merge(self, subgraph, label=None, *property_keys):
        """ Run a :meth:`~py2neo.Transaction.merge` operation within an
        auto-commit :class:`~py2neo.Transaction`.

        The example code below shows a simple merge for a new relationship
        between two new nodes:

            >>> from py2neo import Graph, Node, Relationship
            >>> g = Graph()
            >>> a = Node("Person", name="Alice", age=33)
            >>> b = Node("Person", name="Bob", age=44)
            >>> KNOWS = Relationship.type("KNOWS")
            >>> g.merge(KNOWS(a, b), "Person", "name")

        Following on, we then create a third node (of a different type) to
        which both the original nodes connect:

            >>> c = Node("Company", name="ACME")
            >>> c.__primarylabel__ = "Company"
            >>> c.__primarykey__ = "name"
            >>> WORKS_FOR = Relationship.type("WORKS_FOR")
            >>> g.merge(WORKS_FOR(a, c) | WORKS_FOR(b, c))

        For details of how the merge algorithm works, see the
        :meth:`~py2neo.Transaction.merge` method. Note that this is different
        to a Cypher MERGE.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        :param label: label on which to match any existing nodes
        :param property_keys: property keys on which to match any existing nodes
        """
        self.update(lambda tx: tx.merge(subgraph, label, *property_keys))

    @property
    def nodes(self):
        """ A :class:`.NodeMatcher` for this graph.

        This can be used to find nodes that match given criteria:

            >>> graph = Graph()
            >>> graph.nodes[1234]
            (_1234:Person {name: 'Alice'})
            >>> graph.nodes.get(1234)
            (_1234:Person {name: 'Alice'})
            >>> graph.nodes.match("Person", name="Alice").first()
            (_1234:Person {name: 'Alice'})

        Nodes can also be efficiently counted using this attribute:

            >>> len(graph.nodes)
            55691
            >>> len(graph.nodes.match("Person", age=33))
            12

        """
        return NodeMatcher(self)

    def pull(self, subgraph):
        """ Pull data to one or more entities from their remote counterparts.

        :param subgraph: the collection of nodes and relationships to pull
        """
        self.update(lambda tx: tx.pull(subgraph))

    def push(self, subgraph):
        """ Push data from one or more entities to their remote counterparts.

        :param subgraph: the collection of nodes and relationships to push
        """
        self.update(lambda tx: tx.push(subgraph))

    @property
    def relationships(self):
        """ A :class:`.RelationshipMatcher` for this graph.

        This can be used to find relationships that match given criteria
        as well as efficiently count relationships.
        """
        return RelationshipMatcher(self)

    def separate(self, subgraph):
        """ Run a :meth:`~py2neo.Transaction.separate`
        operation within an auto-commit :class:`~py2neo.Transaction`.

        Note that only relationships which are bound to corresponding
        remote relationships though the ``graph`` and ``identity``
        attributes will trigger a deletion.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        self.update(lambda tx: tx.separate(subgraph))


class SystemGraph(Graph):
    """ A subclass of :class:`.Graph` that provides access to the
    system database for the remote DBMS. This is only available in
    Neo4j 4.0 and above.

    *New in version 2020.0.*
    """

    def __init__(self, profile=None, **settings):
        settings["name"] = "system"
        super(SystemGraph, self).__init__(profile, **settings)

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.service.uri)


class Schema(object):
    """ The schema resource attached to a `Graph` instance.
    """

    def __init__(self, graph):
        self.graph = graph

    @property
    def node_labels(self):
        """ The set of node labels currently defined within the graph.
        """
        return frozenset(record[0] for record in
                         self.graph.run("CALL db.labels"))

    @property
    def relationship_types(self):
        """ The set of relationship types currently defined within the graph.
        """
        return frozenset(record[0] for record in
                         self.graph.run("CALL db.relationshipTypes"))

    def create_index(self, label, *property_keys):
        """ Create a schema index for a label and property
        key combination.
        """
        cypher = "CREATE INDEX ON :{}({})".format(
            cypher_escape(label), ", ".join(map(cypher_escape, property_keys)))
        self.graph.update(cypher)
        while property_keys not in self.get_indexes(label):
            sleep(0.1)

    def create_uniqueness_constraint(self, label, property_key):
        """ Create a node uniqueness constraint for a given label and property
        key.

        While indexes support the use of composite keys, unique constraints may
        only be tied to a single property key.
        """
        cypher = "CREATE CONSTRAINT ON (_:{}) ASSERT _.{} IS UNIQUE".format(
            cypher_escape(label), cypher_escape(property_key))
        self.graph.update(cypher)
        while property_key not in self.get_uniqueness_constraints(label):
            sleep(0.1)

    def drop_index(self, label, *property_keys):
        """ Remove label index for a given property key.
        """
        cypher = "DROP INDEX ON :{}({})".format(
            cypher_escape(label), ", ".join(map(cypher_escape, property_keys)))
        self.graph.update(cypher)

    def drop_uniqueness_constraint(self, label, property_key):
        """ Remove the node uniqueness constraint for a given label and
        property key.
        """
        cypher = "DROP CONSTRAINT ON (_:{}) ASSERT _.{} IS UNIQUE".format(
            cypher_escape(label), cypher_escape(property_key))
        self.graph.update(cypher)

    def _get_indexes(self, label, unique_only=False):
        indexes = []
        result = self.graph.run("CALL db.indexes")
        for record in result:
            properties = []
            # The code branches here depending on the format of the response
            # from the `db.indexes` procedure, which has varied enormously
            # since 3.0.
            if len(record) == 10:
                if "labelsOrTypes" in result.keys():
                    # 4.0.0
                    # ['id', 'name', 'state', 'populationPercent',
                    # 'uniqueness', 'type', 'entityType', 'labelsOrTypes',
                    #  'properties', 'provider']
                    (id_, name, state, population_percent, uniqueness, type_,
                     entity_type, token_names, properties, provider) = record
                    description = None
                    # The 'type' field has randomly changed its meaning in 4.0,
                    # holding for example 'BTREE' instead of for example
                    # 'node_unique_property'. To check for uniqueness, we now
                    # need to look at the new 'uniqueness' field.
                    is_unique = uniqueness == "UNIQUE"
                else:
                    # 3.5.3
                    # ['description', 'indexName', 'tokenNames', 'properties',
                    #  'state', 'type', 'progress', 'provider', 'id',
                    #  'failureMessage']
                    (description, index_name, token_names, properties, state,
                     type_, progress, provider, id_, failure_message) = record
                    is_unique = type_ == "node_unique_property"
            elif len(record) == 7:
                # 3.4.10
                (description, lbl, properties, state,
                 type_, provider, failure_message) = record
                is_unique = type_ == "node_unique_property"
                token_names = [lbl]
            elif len(record) == 6:
                # 3.4.7
                description, lbl, properties, state, type_, provider = record
                is_unique = type_ == "node_unique_property"
                token_names = [lbl]
            elif len(record) == 3:
                # 3.0.10
                description, state, type_ = record
                is_unique = type_ == "node_unique_property"
                token_names = []
            else:
                raise RuntimeError("Unexpected response from procedure "
                                   "db.indexes (%d fields)" % len(record))
            if state not in (u"ONLINE", u"online"):
                continue
            if unique_only and not is_unique:
                continue
            if not token_names or not properties:
                if description:
                    from py2neo.cypher.lexer import CypherLexer
                    from pygments.token import Token
                    tokens = list(CypherLexer().get_tokens(description))
                    for token_type, token_value in tokens:
                        if token_type is Token.Name.Label:
                            token_names.append(token_value.strip("`"))
                        elif token_type is Token.Name.Variable:
                            properties.append(token_value.strip("`"))
            if not token_names or not properties:
                continue
            if label in token_names:
                indexes.append(tuple(properties))
        return indexes

    def get_indexes(self, label):
        """ Fetch a list of indexed property keys for a label.
        """
        return self._get_indexes(label)

    def get_uniqueness_constraints(self, label):
        """ Fetch a list of unique constraints for a label. Each constraint is
        the name of a single property key.
        """
        return [k[0] for k in self._get_indexes(label, unique_only=True)]


class Transaction(object):
    """ Logical context for one or more graph operations.

    Transaction objects are typically constructed by the
    :meth:`.Graph.auto` and :meth:`.Graph.begin` methods.
    Likewise, the :meth:`.Graph.commit` and :meth:`.Graph.rollback`
    methods can be used to finish a transaction.
    """

    def __init__(self, graph, autocommit=False, readonly=False,
                 # after=None, metadata=None, timeout=None
                 ):
        self._graph = graph
        self._autocommit = autocommit
        self._connector = self.graph.service.connector
        if autocommit:
            self._ref = None
        else:
            self._ref = self._connector.begin(self.graph.name, readonly=readonly,
                                              # after, metadata, timeout
                                              )
        self._readonly = readonly
        self._closed = False
        self._bookmark = None
        self._profile = None
        self._time = None

    @property
    def graph(self):
        """ Graph to which this transaction belongs.
        """
        return self._graph

    @property
    def ref(self):
        """ Transaction reference.
        """
        return self._ref

    @property
    def readonly(self):
        """ :py:const:`True` if this is a readonly transaction,
        :py:const:`False` otherwise.
        """
        return self._readonly

    @property
    def closed(self):
        """ :py:const:`True` if this transaction is closed,
        :py:const:`False` otherwise.
        """
        return self._closed

    @property
    def bookmark(self):
        """ The closing bookmark for this transaction, populated
        on commit.
        """
        return self._bookmark

    @property
    def profile(self):
        """ The connection profile under which this transaction was
        carried out.
        """
        return self._profile

    @property
    def time(self):
        """ The total time taken to carry out this transaction
        """
        return self._time

    def run(self, cypher, parameters=None, **kwparameters):
        """ Send a Cypher query to the server for execution and return
        a :py:class:`~.cypher.Cursor` for navigating its result.

        :param cypher: Cypher query
        :param parameters: dictionary of parameters
        :returns: :py:class:`~.cypher.Cursor` object
        """
        from py2neo.client import Connection

        if self.closed:
            raise TypeError("Cannot run query in closed transaction")

        try:
            hydrant = Connection.default_hydrant(self._connector.profile, self.graph)
            parameters = dict(parameters or {}, **kwparameters)
            if self.ref:
                result = self._connector.run(self.ref, cypher, parameters)
            else:
                result = self._connector.auto_run(cypher, parameters,
                                                  graph_name=self.graph.name,
                                                  readonly=self.readonly)
            self._connector.pull(result, -1)
            return Cursor(result, hydrant)
        finally:
            if not self.ref:
                self._closed = True

    def evaluate(self, cypher, parameters=None, **kwparameters):
        """ Execute a single Cypher query and return the value from
        the first column of the first record.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        :returns: single return value or :const:`None`
        """
        return self.run(cypher, parameters, **kwparameters).evaluate(0)

    def update(self, cypher, parameters=None, **kwparameters):
        """ Execute a single Cypher statement and discard any result
        returned.

        :param cypher: Cypher statement
        :param parameters: dictionary of parameters
        """
        self.run(cypher, parameters, **kwparameters)

    @deprecated("The transaction.commit() method is deprecated, "
                "use graph.commit(transaction) instead")
    def commit(self):
        """ Commit the transaction.
        """
        return self.graph.commit(self)

    @deprecated("The transaction.rollback() method is deprecated, "
                "use graph.rollback(transaction) instead")
    def rollback(self):
        """ Roll back the current transaction, undoing all actions
        previously taken.
        """
        return self.graph.rollback(self)

    def create(self, subgraph):
        """ Create remote nodes and relationships that correspond to those in a
        local subgraph. Any entities in *subgraph* that are already bound to
        remote entities will remain unchanged, those which are not will become
        bound to their newly-created counterparts.

        For example::

            >>> from py2neo import Graph, Node, Relationship
            >>> g = Graph()
            >>> tx = g.begin()
            >>> a = Node("Person", name="Alice")
            >>> tx.create(a)
            >>> b = Node("Person", name="Bob")
            >>> ab = Relationship(a, "KNOWS", b)
            >>> tx.create(ab)
            >>> tx.commit()
            >>> g.exists(ab)
            True

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                    creatable object
        """
        if self._autocommit:
            raise TypeError("Create operations are not supported inside "
                            "auto-commit transactions")
        try:
            create = subgraph.__db_create__
        except AttributeError:
            raise TypeError("No method defined to create object %r" % subgraph)
        else:
            create(self)

    def delete(self, subgraph):
        """ Delete the remote nodes and relationships that correspond to
        those in a local subgraph. To delete only the relationships, use
        the :meth:`.separate` method.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            delete = subgraph.__db_delete__
        except AttributeError:
            raise TypeError("No method defined to delete object %r" % subgraph)
        else:
            delete(self)

    def exists(self, subgraph):
        """ Determine whether one or more entities all exist within the
        graph. Note that if any nodes or relationships in *subgraph* are not
        bound to remote counterparts, this method will return ``False``.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        :returns: ``True`` if all entities exist remotely, ``False`` otherwise
        """
        try:
            exists = subgraph.__db_exists__
        except AttributeError:
            raise TypeError("No method defined to check existence of object %r" % subgraph)
        else:
            return exists(self)

    def merge(self, subgraph, primary_label=None, primary_key=None):
        """ Create or update the nodes and relationships of a local
        subgraph in the remote database. Note that the functionality of
        this operation is not strictly identical to the Cypher MERGE
        clause, although there is some overlap.

        Each node and relationship in the local subgraph is merged
        independently, with nodes merged first and relationships merged
        second.

        For each node, the merge is carried out by comparing that node with
        a potential remote equivalent on the basis of a single label and
        property value. If no remote match is found, a new node is created;
        if a match is found, the labels and properties of the remote node
        are updated. The label and property used for comparison are determined
        by the `primary_label` and `primary_key` arguments but may be
        overridden for individual nodes by the of `__primarylabel__` and
        `__primarykey__` attributes on the node itself.

        For each relationship, the merge is carried out by comparing that
        relationship with a potential remote equivalent on the basis of matching
        start and end nodes plus relationship type. If no remote match is found,
        a new relationship is created; if a match is found, the properties of
        the remote relationship are updated.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph` object
        :param primary_label: label on which to match any existing nodes
        :param primary_key: property key(s) on which to match any existing
                            nodes
        """
        try:
            merge = subgraph.__db_merge__
        except AttributeError:
            raise TypeError("No method defined to merge object %r" % subgraph)
        else:
            merge(self, primary_label, primary_key)

    def pull(self, subgraph):
        """ Update local entities from their remote counterparts.

        For any nodes and relationships that exist in both the local
        :class:`.Subgraph` and the remote :class:`.Graph`, pull properties
        and node labels into the local copies. This operation does not
        create or delete any entities.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            pull = subgraph.__db_pull__
        except AttributeError:
            raise TypeError("No method defined to pull object %r" % subgraph)
        else:
            return pull(self)

    def push(self, subgraph):
        """ Update remote entities from their local counterparts.

        For any nodes and relationships that exist in both the local
        :class:`.Subgraph` and the remote :class:`.Graph`, push properties
        and node labels into the remote copies. This operation does not
        create or delete any entities.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            push = subgraph.__db_push__
        except AttributeError:
            raise TypeError("No method defined to push object %r" % subgraph)
        else:
            return push(self)

    def separate(self, subgraph):
        """ Delete the remote relationships that correspond to those in a local
        subgraph. This leaves any nodes untouched.

        :param subgraph: a :class:`.Node`, :class:`.Relationship` or other
                       :class:`.Subgraph`
        """
        try:
            separate = subgraph.__db_separate__
        except AttributeError:
            raise TypeError("No method defined to separate object %r" % subgraph)
        else:
            separate(self)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright Nigel Small
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import, division

from codecs import decode
from collections import OrderedDict
from datetime import date, time, datetime, timedelta
from io import BytesIO
from struct import pack as struct_pack, unpack as struct_unpack

from pytz import FixedOffset, timezone, utc
import six

from interchange.geo import Point
from interchange.time import Duration, Date, Time, DateTime, UnixEpoch


INT_DATA = {}
for n in range(-0x8000, 0x8000):
    if -0x10 <= n < 0x80:
        INT_DATA[n] = struct_pack(">b", n)
    elif -0x80 <= n < -0x10:
        INT_DATA[n] = b"\xC8" + struct_pack(">b", n)
    else:
        INT_DATA[n] = b"\xC9" + struct_pack(">h", n)

BYTES_S_HEAD = [b"\xCC" + struct_pack(">B", value) for value in range(0x100)]
BYTES_M_HEAD = [b"\xCD" + struct_pack(">H", value) for value in range(0x10000)]

STR_S_HEAD = [b"\xD0" + struct_pack(">B", value) for value in range(0x100)]
STR_M_HEAD = [b"\xD1" + struct_pack(">H", value) for value in range(0x10000)]

LIST_S_HEAD = [b"\xD4" + struct_pack(">B", value) for value in range(0x100)]
LIST_M_HEAD = [b"\xD5" + struct_pack(">H", value) for value in range(0x10000)]

DICT_S_HEAD = [b"\xD8" + struct_pack(">B", value) for value in range(0x100)]
DICT_M_HEAD = [b"\xD9" + struct_pack(">H", value) for value in range(0x10000)]

UNPACKED_UINT_8 = {bytes(bytearray([x])): x for x in range(0x100)}
UNPACKED_UINT_16 = {struct_pack(">H", x): x for x in range(0x10000)}

UNPACKED_MARKERS = {b"\xC0": None, b"\xC2": False, b"\xC3": True}
UNPACKED_MARKERS.update({bytes(bytearray([z])): z for z in range(0x00, 0x80)})
UNPACKED_MARKERS.update({bytes(bytearray([z + 256])): z for z in range(-0x10, 0x00)})


INT64_MIN = -(2 ** 63)
INT64_MAX = 2 ** 63


UNIX_EPOCH_DATE_ORDINAL = UnixEpoch.to_ordinal()


class Structure(object):

    def __init__(self, tag, *fields):
        self.tag = tag
        self.fields = list(fields)

    def __repr__(self):
        return "Structure[#%02X](%s)" % (self.tag, ", ".join(map(repr, self.fields)))

    def __eq__(self, other):
        try:
            return self.tag == other.tag and self.fields == other.fields
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.fields)

    def __getitem__(self, key):
        return self.fields[key]

    def __setitem__(self, key, value):
        self.fields[key] = value


class Packer(object):

    integer_types = six.integer_types
    text_type = six.text_type
    bytearray_type = bytearray
    list_type = list
    dict_type = dict

    def __init__(self, buffer, version=()):
        self._buffer = buffer
        self._write = self._buffer.write
        self.version = version

    def pack(self, value):
        t = type(value)

        # String (Unicode)
        if t is self.text_type:
            self._pack_unicode(value)

        # List
        elif t is self.list_type or t is tuple:
            self._pack_list(value)

        # Integer
        elif t in self.integer_types:
            self._pack_integer(value)

        # Float
        elif t is float:
            self._write(b"\xC1")
            self._write(struct_pack(">d", value))

        # Boolean
        elif value is True:
            self._write(b"\xC3")
        elif value is False:
            self._write(b"\xC2")

        # None
        elif value is None:
            self._write(b"\xC0")  # NULL

        # Byte array
        elif t is self.bytearray_type:
            self._pack_bytearray(value)

        # String (UTF-8)
        elif t is bytes:
            self._pack_utf8(value)

        # Dictionary
        elif t is dict or t is OrderedDict or isinstance(value, self.dict_type):
            self._pack_dict(value)

        # Bolt 2 introduced temporal and spatial types
        elif self.version < (2, 0):
            raise TypeError("Values of type %s are not supported "
                            "by Bolt %s" % (type(value), ".".join(self.version)))

        # DateTime
        #
        # Note: The built-in datetime.datetime class extends the
        # datetime.date class, so this needs to be listed first
        # to avoid objects being encoded incorrectly.
        #
        elif isinstance(value, datetime):
            self._pack_datetime(DateTime.from_native(value))
        elif isinstance(value, DateTime):
            self._pack_datetime(value)

        # Date
        elif isinstance(value, (date, Date)):
            self._write(b"\xB1D")
            self._pack_integer(value.toordinal() - UNIX_EPOCH_DATE_ORDINAL)

        # Time
        elif isinstance(value, (time, Time)):
            self._pack_time(value)

        # TimeDelta
        elif isinstance(value, timedelta):
            self._write(b"\xB4E")
            self._pack_integer(0)                                   # months
            self._pack_integer(value.days)                          # days
            self._pack_integer(value.seconds)                       # seconds
            self._pack_integer(1000 * value.microseconds)           # nanoseconds

        # Duration
        elif isinstance(value, Duration):
            self._write(b"\xB4E")
            self._pack_integer(value.months)                        # months
            self._pack_integer(value.days)                          # days
            self._pack_integer(value.seconds)                       # seconds
            self._pack_integer(int(1000000000 * value.subseconds))  # nanoseconds

        # Point
        elif isinstance(value, Point):
            self._pack_point(value)

        # Other
        else:
            raise TypeError("Values of type %s are not supported" % type(value))

    def _pack_unicode(self, value):
        # Count the number of chars and handle boundary cases
        size = len(value)
        if size == 0:
            self._write(b"\x80")
            return
        elif size >= 0x100000000:
            raise ValueError("String too large")
        # Count the number of bytes when encoded as UTF-8
        self._pack_utf8(value.encode("utf-8"))

    def _pack_utf8(self, value):
        size = len(value)
        # Write the string header
        if size < 0x10:
            self._write(bytearray([0x80 + size]))
        elif size < 0x100:
            self._write(STR_S_HEAD[size])
        elif size < 0x10000:
            self._write(STR_M_HEAD[size])
        elif size < 0x100000000:
            self._write(b"\xD2")
            self._write(struct_pack(">I", size))
        else:
            raise ValueError("String too large")
        # Write the string content
        self._write(value)

    def _pack_list(self, value):
        size = len(value)
        if size == 0:
            self._write(b"\x90")
            return
        elif size < 0x10:
            self._write(bytearray([0x90 + size]))
        elif size < 0x100:
            self._write(LIST_S_HEAD[size])
        elif size < 0x10000:
            self._write(LIST_M_HEAD[size])
        elif size < 0x100000000:
            self._write(b"\xD6")
            self._write(struct_pack(">I", size))
        else:
            raise ValueError("List too large")
        for item in value:
            self.pack(item)

    def _pack_integer(self, value):
        if -0x8000 <= value < 0x8000:
            self._write(INT_DATA[value])
        elif -0x80000000 <= value < 0x80000000:
            self._write(b"\xCA")
            self._write(struct_pack(">i", value))
        elif INT64_MIN <= value < INT64_MAX:
            self._write(b"\xCB")
            self._write(struct_pack(">q", value))
        else:
            raise ValueError("Integer %s out of range" % value)

    def _pack_dict(self, value):
        size = len(value)
        if size == 0:
            self._write(b"\xA0")
        elif size < 0x10:
            self._write(bytearray([0xA0 + size]))
        elif size < 0x100:
            self._write(DICT_S_HEAD[size])
        elif size < 0x10000:
            self._write(DICT_M_HEAD[size])
        elif size < 0x100000000:
            self._write(b"\xDA")
            self._write(struct_pack(">I", size))
        else:
            raise ValueError("Dictionary too large")
        for key, item in value.items():
            t = type(key)
            if t is self.text_type:
                self._pack_unicode(key)
            elif t is bytes:
                self._pack_utf8(key)
            else:
                raise TypeError("Dictionary keys must be "
                                "of type %r or bytes" % self.text_type)
            self.pack(item)

    def _pack_bytearray(self, value):
        size = len(value)
        if size < 0x100:
            self._write(BYTES_S_HEAD[size])
        elif size < 0x10000:
            self._write(BYTES_M_HEAD[size])
        elif size < 0x100000000:
            self._write(b"\xCE")
            self._write(struct_pack(">I", size))
        else:
            raise ValueError("Byte array too large")
        self._write(value)

    def _pack_time(self, t):
        try:
            nanoseconds = int(t.ticks * 1000000000)
        except AttributeError:
            nanoseconds = (3600000000000 * t.hour + 60000000000 * t.minute +
                           1000000000 * t.second + 1000 * t.microsecond)
        if t.tzinfo:
            self._write(b"\xB2T")
            self._pack_integer(nanoseconds)
            self._pack_integer(t.tzinfo.utcoffset(t).seconds)
        else:
            self._write(b"\xB1t")
            self._pack_integer(nanoseconds)

    def _pack_datetime(self, dt):
        tz = dt.tzinfo
        if tz is None:
            # without time zone
            seconds, nanoseconds = utc.localize(dt).to_clock_time() - UnixEpoch.to_clock_time()
            self._write(b"\xB2d")
            self._pack_integer(seconds)
            self._pack_integer(nanoseconds)
        elif hasattr(tz, "zone") and tz.zone:
            # with named time zone
            zone_epoch = DateTime(1970, 1, 1, tzinfo=dt.tzinfo)
            seconds, nanoseconds = dt.to_clock_time() - zone_epoch.to_clock_time()
            self._write(b"\xB3f")
            self._pack_integer(seconds)
            self._pack_integer(nanoseconds)
            self.pack(tz.zone)
        else:
            # with time offset
            zone_epoch = DateTime(1970, 1, 1, tzinfo=dt.tzinfo)
            seconds, nanoseconds = dt.to_clock_time() - zone_epoch.to_clock_time()
            self._write(b"\xB3F")
            self._pack_integer(seconds)
            self._pack_integer(nanoseconds)
            self._pack_integer(tz.utcoffset(dt).seconds)

    def _pack_point(self, p):
        dim = len(p)
        self._write(bytearray([0xB1 + dim]))
        if dim == 2:
            self._write(b"X")
        elif dim == 3:
            self._write(b"Y")
        else:
            raise ValueError("Cannot dehydrate Point with %d dimensions" % dim)
        self._pack_integer(p.srid)
        for value in p:
            self.pack(value)


class Unpacker(object):

    def __init__(self, data, offset=0):
        if six.PY2:
            self._data = bytearray(data)  # FIXME
        else:
            self._data = data
        self._offset = offset

    def unpack(self):
        marker = self._data[self._offset]
        self._offset += 1

        # Tiny collections
        if marker == 0x80:
            return ""
        elif 0x81 <= marker <= 0x8F:  # TINY_STRING
            start = self._offset
            self._offset = start + (marker & 0x0F)
            return decode(self._data[start:self._offset], "utf-8")
        elif marker == 0x90:
            return []
        elif marker == 0x91:
            return [self.unpack()]
        elif marker == 0x92:
            return [self.unpack(), self.unpack()]
        elif marker == 0x93:
            return [self.unpack(), self.unpack(), self.unpack()]
        elif 0x94 <= marker <= 0x9F:    # TINY_LIST
            return [self.unpack() for _ in range(marker & 0x0F)]
        elif 0xA0 <= marker <= 0xAF:    # TINY_DICT
            size = marker & 0x0F
            value = {}
            for _ in range(size):
                key = self.unpack()
                value[key] = self.unpack()
            return value

        # Integer
        elif 0x00 <= marker <= 0x7F:
            return marker
        elif 0xF0 <= marker <= 0xFF:
            return marker - 0x100
        elif marker == 0xC8:
            return self._read_i8()
        elif marker == 0xC9:
            return self._read_i16be()
        elif marker == 0xCA:
            return self._read_i32be()
        elif marker == 0xCB:
            return self._read_i64be()

        # String
        elif marker == 0xD0:  # STRING_8:
            size = self._read_u8()
            return decode(self._read(size), "utf-8")
        elif marker == 0xD1:  # STRING_16:
            size = self._read_u16be()
            return decode(self._read(size), "utf-8")
        elif marker == 0xD2:  # STRING_32:
            size = self._read_u32be()
            return decode(self._read(size), "utf-8")

        # Structure
        elif 0xB0 <= marker <= 0xBF:    # TINY_STRUCT
            tag = self._read_u8()
            fields = [self.unpack() for _ in range(marker & 0x0F)]
            if tag == 68:  # 'D'
                return self._hydrate_date(*fields)
            elif tag in (84, 116):  # 'T' and 't'
                return self._hydrate_time(*fields)
            elif tag in (70, 100, 102):  # b"F", b"f", b"d"
                return self._hydrate_datetime(*fields)
            elif tag == 69:  # b"E"
                return self._hydrate_duration(*fields)
            elif tag in (88, 89):  # b"X", b"Y"
                return self._hydrate_point(*fields)
            else:
                return Structure(tag, *fields)

        # Float
        elif marker == 0xC1:
            return self._read_f64be()

        # Boolean
        elif marker == 0xC2:
            return False
        elif marker == 0xC3:
            return True

        # Null
        elif marker == 0xC0:
            return None

        # Bytes
        elif marker == 0xCC:
            size = self._read_u8()
            return bytes(self._read(size))
        elif marker == 0xCD:
            size = self._read_u16be()
            return bytes(self._read(size))
        elif marker == 0xCE:
            size = self._read_u32be()
            return bytes(self._read(size))

        # List
        elif marker == 0xD4:  # LIST_8:
            size = self._read_u8()
            return [self.unpack() for _ in range(size)]
        elif marker == 0xD5:  # LIST_16:
            size = self._read_u16be()
            return [self.unpack() for _ in range(size)]
        elif marker == 0xD6:  # LIST_32:
            size = self._read_u32be()
            return [self.unpack() for _ in range(size)]

        # Dictionary
        elif marker == 0xD8:  # MAP_8:
            size = self._read_u8()
            value = {}
            for _ in range(size):
                key = self.unpack()
                value[key] = self.unpack()
            return value
        elif marker == 0xD9:  # MAP_16:
            size = self._read_u16be()
            value = {}
            for _ in range(size):
                key = self.unpack()
                value[key] = self.unpack()
            return value
        elif marker == 0xDA:  # MAP_32:
            size = self._read_u32be()
            value = {}
            for _ in range(size):
                key = self.unpack()
                value[key] = self.unpack()
            return value

        else:
            raise ValueError("Unknown PackStream marker %02X" % marker)

    def _hydrate_date(self, days):
        """ Hydrator for `Date` values.

        :param days:
        :return: Date
        """
        return Date.from_ordinal(UNIX_EPOCH_DATE_ORDINAL + days)

    def _hydrate_time(self, nanoseconds, tz=None):
        """ Hydrator for `Time` and `LocalTime` values.

        :param nanoseconds:
        :param tz:
        :return: Time
        """
        seconds, nanoseconds = map(int, divmod(nanoseconds, 1000000000))
        minutes, seconds = map(int, divmod(seconds, 60))
        hours, minutes = map(int, divmod(minutes, 60))
        seconds = (1000000000 * seconds + nanoseconds) / 1000000000
        t = Time(hours, minutes, seconds)
        if tz is None:
            return t
        tz_offset_minutes, tz_offset_seconds = divmod(tz, 60)
        zone = FixedOffset(tz_offset_minutes)
        return zone.localize(t)

    def _hydrate_datetime(self, seconds, nanoseconds, tz=None):
        """ Hydrator for `DateTime` and `LocalDateTime` values.

        :param seconds:
        :param nanoseconds:
        :param tz:
        :return: datetime
        """
        minutes, seconds = map(int, divmod(seconds, 60))
        hours, minutes = map(int, divmod(minutes, 60))
        days, hours = map(int, divmod(hours, 24))
        seconds = (1000000000 * seconds + nanoseconds) / 1000000000
        t = DateTime.combine(Date.from_ordinal(UNIX_EPOCH_DATE_ORDINAL + days),
                             Time(hours, minutes, seconds))
        if tz is None:
            return t
        if isinstance(tz, int):
            tz_offset_minutes, tz_offset_seconds = divmod(tz, 60)
            zone = FixedOffset(tz_offset_minutes)
        else:
            zone = timezone(tz)
        return zone.localize(t)

    def _hydrate_duration(self, months, days, seconds, nanoseconds):
        """ Hydrator for `Duration` values.

        :param months:
        :param days:
        :param seconds:
        :param nanoseconds:
        :return: `duration` namedtuple
        """
        return Duration(months=months, days=days, seconds=seconds, nanoseconds=nanoseconds)

    def _hydrate_point(self, srid, *coordinates):
        """ Create a new instance of a Point subclass from a raw
        set of fields. The subclass chosen is determined by the
        given SRID; a ValueError will be raised if no such
        subclass can be found.
        """
        try:
            point_class, dim = Point.class_for_srid(srid)
        except KeyError:
            point = Point(coordinates)
            point.srid = srid
            return point
        else:
            if len(coordinates) != dim:
                raise ValueError("SRID %d requires %d coordinates (%d provided)" % (srid, dim, len(coordinates)))
            return point_class(coordinates)

    def _read(self, n=1):
        q = self._offset + n
        m = self._data[self._offset:q]
        self._offset = q
        return m

    def _read_u8(self):
        q = self._offset + 1
        n, = struct_unpack(">B", self._data[self._offset:q])
        self._offset = q
        return n

    def _read_u16be(self):
        q = self._offset + 2
        n, = struct_unpack(">H", self._data[self._offset:q])
        self._offset = q
        return n

    def _read_u32be(self):
        q = self._offset + 4
        n, = struct_unpack(">I", self._data[self._offset:q])
        self._offset = q
        return n

    def _read_i8(self):
        q = self._offset + 1
        z, = struct_unpack(">b", self._data[self._offset:q])
        self._offset = q
        return z

    def _read_i16be(self):
        q = self._offset + 2
        z, = struct_unpack(">h", self._data[self._offset:q])
        self._offset = q
        return z

    def _read_i32be(self):
        q = self._offset + 4
        z, = struct_unpack(">i", self._data[self._offset:q])
        self._offset = q
        return z

    def _read_i64be(self):
        q = self._offset + 8
        z, = struct_unpack(">q", self._data[self._offset:q])
        self._offset = q
        return z

    def _read_f64be(self):
        q = self._offset + 8
        r, = struct_unpack(">d", self._data[self._offset:q])
        self._offset = q
        return r


def pack(*values, **kwargs):
    buffer = BytesIO()
    packer = Packer(buffer, **kwargs)
    for value in values:
        packer.pack(value)
    return buffer.getvalue()


def unpack(data, offset=0):
    s = Unpacker(data, offset)
    while True:
        try:
            yield s.unpack()
        except IndexError:
            break
